/*

   Copyright 2012-2013 SRI International

   Licensed under the Apache License, Version 2.0 (the "License"); you may
   not use this file except in compliance with the License. You may obtain
   a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

   The development was supported by National Science Foundation Grant 0932397:
   A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.app.brainmap;

//import android.util.Log;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

import cf.api.*;
import cf.app.brainmap.GraphKnowledge.TYPE;
import cf.app.brainmap.events.*;
import cf.common.Param;
import cf.real.IPKnowledge;
import cf.util.Dist;


/*
	  BrainmapApp is the main cyber-application that mediates the events and
	knowledge flow. It handles ShapeEvent containing geometry data created when
	a local user draws a diagram on a map. BrainmapApp generates ShapeKnowledge sk
	and corresponding GraphKnowledge gk that are posted as DisseminationEvent 
	(DEvent), which will be later handled by SketchApp and GraphApp, respectively. 
	  This pairing between sk and gk is done by assigning same uuid to enable 
	composing a node or an edge on the map from sk and gk that are separately 
	delivered.
 */
public class BrainmapApp extends CyberApp implements WaitingQueueListener {
	
	// Delay for posting Session Beacon Knowledge
	private static double SESSION_BEACON_DELAY = Param.HelloBroadcastDelay*3;	//s
	
	// Session Time Out
	private static double SESSION_TIMEOUT_DELAY = SESSION_BEACON_DELAY*1.5;		//s
	
	// Delay for checking Sessions's 
	private static double SESSION_CHECK_DELAY = SESSION_TIMEOUT_DELAY;			//s
			
	// Queues for paring ShapeKnowledge and GraphKnowledge
	private WaitingQueue shapeKnowledgeWaitingQueue;
	private WaitingQueue graphKnowledgeWaitingQueue;
	
	// For Session Control
	private ArrayList<Session> sessions;
	private Hashtable<String, Long> sessionBeaconTimeTable;
	private Session mySession;
	
	private AppKnowledgeBase appKB;

	public BrainmapApp(CyberNode node) {
		super(node);
		initQueues();
		
		mySession = null;
		sessions = new ArrayList<Session>();
		sessionBeaconTimeTable = new Hashtable<String, Long>();
		
		initAppKB();
	}

	@Override
	public String name() {
		return super.name() + "::BrainmapApp";
	}

	@Override
	public void initialize() {
		postEvent(new SessionEvent(SessionEvent.TYPE.BEACON, null, false, SESSION_BEACON_DELAY, this));
		postEvent(new SessionEvent(SessionEvent.TYPE.CHECK, null, false, SESSION_CHECK_DELAY, this));
	}
	
	@Override
	public void handleKnowledge(ApplicationKnowledge k){
		if(k instanceof SessionKnowledge || k instanceof SessionBeaconKnowledge)
			appKB.updateKnowledge(k);	
	}
	
	@Override
	public void handleEvent(ApplicationEvent e) {
		
		if (e instanceof ShapeEvent)
			handleShapeInputEvent((ShapeEvent)e);
		
		/*else if (e instanceof DisseminationEvent)	
			handleDisseminationEvent((DisseminationEvent) e);*/
		else if(e instanceof KnowledgeAcceptanceEvent)
			handleKnowledgeAcceptanceEvent((KnowledgeAcceptanceEvent) e);
		
		else if(e instanceof SessionEvent)
			handleSessionEvent((SessionEvent)e);
	}
	
	private void handleSessionEvent(SessionEvent e){
		switch(e.type){
		case CHECK:
			checkSessions();
			break;
			
		case BEACON:
			sendSessionBeacon();
			break;
			
		case REQUEST:
			handleSessionRequestEvent();
			break;
			
		case CREATE:
			createSession(e.session);
			break;
			
		case EXIT:
			exitSession();
			break;
			
		case JOIN:
			joinSession(e.session);
			break;
			
		case LOAD:
			loadSession(e.session);
			break;
		}
	}
	
	private void handleSessionRequestEvent(){
		ArrayList<Session> response = new ArrayList<Session>();
		
		for(Session s : sessions){
			if(null != sessionBeaconTimeTable.get(s.id))
				response.add(s);
		}
		
		postEvent(new SessionEvent(response));
	}
	
	private void loadSession(Session s){
		if(null == sessionBeaconTimeTable.get(s.id)){
			postEvent(new SessionEvent(SessionEvent.TYPE.LOAD, s, true, 0.0, BrainmapRealWorld.sketchApp));
			mySession = s;
			
			Session newSession = s.clone();
			newSession.addEngine(BrainmapRealEngine.myEngineId());
			postKnowledge(new SessionKnowledge(newSession));
		}
		
		else
			postEvent(new SessionEvent(SessionEvent.TYPE.LOAD, getSession(s.id), false, 0.0, BrainmapRealWorld.sketchApp));
	}
	
	private void joinSession(Session s){
		mySession = s;
		postEvent(new SessionEvent(SessionEvent.TYPE.JOIN, s, false, 0.0, BrainmapRealWorld.sketchApp));
		
		Session newSession = s.clone();
		newSession.addEngine(BrainmapRealEngine.myEngineId());
		postKnowledge(new SessionKnowledge(newSession));
	}
	
	synchronized private boolean containsSession(String sessionId){
		for(Session s : sessions){
			if(s.id.equals(sessionId))
				return true;
		}
		return false;
	}
	
	synchronized private boolean hasSession(Session session){
		for(Session s : sessions){
			if(s.id.equals(session.id)){
				if(s.engines.containsAll(session.engines) && session.engines.containsAll(s.engines))
					return true;
				else 
					return false;
			}
		}
		return false;
	}
	
	synchronized public void removeSession(String id){
		if(null != sessionBeaconTimeTable.get(id))
			sessionBeaconTimeTable.remove(id);
		
		if(containsSession(id)){
			for(int i=0; i<sessions.size(); i++){
				if(sessions.get(i).id.equals(id))
					sessions.remove(i);
			}
		}
	}
	
	synchronized private Session getSession(String sessionId){
		for(int i=0; i<sessions.size(); i++){
			Session s = sessions.get(i);
			if(s.id.equals(sessionId))
				return s;
		}
		
		return null;
	}
	
	private void handleSessionKnowledge(SessionKnowledge k){
		boolean isModified = false;
		
		if(null != mySession && mySession.id.equals(k.id)){
			//System.out.println("Set as my Sessions!!! " + k);
			mySession = k.session;
		}
		
		sessions.add(k.session);
				
		// Update sessions beacon time
		if(null != sessionBeaconTimeTable.get(k.id))
			sessionBeaconTimeTable.remove(k.id);
		sessionBeaconTimeTable.put(k.id, System.currentTimeMillis());
		
		postEvent(new SessionEvent(SessionEvent.TYPE.REQUEST, this));
	}
	
	private void handleSessionBeaconKnowledge(SessionBeaconKnowledge k){
		//System.out.println("Received-Brainmapapp SessionBeaconKnowledge " + k);
		
		if(null != mySession && k.id.equals(mySession.id))
			return;
		
		if(null != sessionBeaconTimeTable.get(k.id)){
			sessionBeaconTimeTable.remove(k.id);
			sessionBeaconTimeTable.put(k.id, System.currentTimeMillis());
		}
		
		else{
			sessionBeaconTimeTable.put(k.id, System.currentTimeMillis());
			postEvent(new SessionEvent(SessionEvent.TYPE.REQUEST, this));
		}
	}
	
	private void sendSessionBeacon(){
		if(null == mySession){
			postEvent(new SessionEvent(SessionEvent.TYPE.BEACON, null, false, SESSION_BEACON_DELAY, this));
			return;
		}
		
		double beaconProbability = (0==mySession.size())? 1.0:1.0/(double)mySession.size();
		double randomval = Dist.dist.sampleUniform(0.0, 1.0);
		
		if(beaconProbability > randomval)
			postKnowledge(new SessionBeaconKnowledge(mySession.id));
		
		postEvent(new SessionEvent(SessionEvent.TYPE.BEACON, null, false, SESSION_BEACON_DELAY, this));
	}
	
	synchronized private void checkSessions(){
		if(null != mySession){
			postEvent(new SessionEvent(SessionEvent.TYPE.CHECK, null, false, SESSION_CHECK_DELAY, this));
			return;
		}
		
		boolean sessionListModified = false;
		long now = System.currentTimeMillis();
		
		Enumeration<String> e = sessionBeaconTimeTable.keys();
		
		while(e.hasMoreElements()){
			String sessionId = e.nextElement();
			long lastBeaconTime = sessionBeaconTimeTable.get(sessionId);			
			if(now - lastBeaconTime > SESSION_TIMEOUT_DELAY*1000){
					sessionBeaconTimeTable.remove(sessionId);
					sessionListModified = true;
			}
		}
		postEvent(new SessionEvent(SessionEvent.TYPE.CHECK, null, false, SESSION_CHECK_DELAY, this));
		
		if(sessionListModified)
			postEvent(new SessionEvent(SessionEvent.TYPE.REQUEST, this));
	}
	
	private void createSession(Session session){
		mySession = session;
		
		postEvent(new SessionEvent(SessionEvent.TYPE.CREATE, mySession, false, 0.0, BrainmapRealWorld.sketchApp));
		postKnowledge(new SessionKnowledge(mySession));
	}
	
	synchronized private void exitSession(){
		Session s = mySession.clone();
		s.removeEngine(BrainmapRealEngine.myEngineId());
		postKnowledge(new SessionKnowledge(s));
		
		mySession = null;
		postEvent(new SessionEvent(SessionEvent.TYPE.SHOW, BrainmapRealWorld.sketchApp));
	}
	
	//private void handleDisseminationEvent(DisseminationEvent e){
	private void handleKnowledgeAcceptanceEvent(KnowledgeAcceptanceEvent e){
		if(((KnowledgeAcceptanceEvent) e).k instanceof ShapeKnowledge){
			ShapeKnowledge sk = (ShapeKnowledge)((KnowledgeAcceptanceEvent) e).k;	
			if(null != mySession && sk.sessionId.equals(mySession.id))
				shapeKnowledgeWaitingQueue.add(sk);	// Paring with GraphKnowledge
		}
		
		else if(((KnowledgeAcceptanceEvent) e).k instanceof GraphKnowledge)
			graphKnowledgeWaitingQueue.add((GraphKnowledge)((KnowledgeAcceptanceEvent) e).k);	//Pairing with ShapeKnowledge
	}
	
	private void handleShapeInputEvent(ShapeEvent e){
		if(null == mySession)
			return;
		
		ShapeKnowledge sk = new ShapeKnowledge(e.s, mySession.id);

		if (e.s instanceof FileContentShape) {
			postEvent(new KnowledgeDisseminationEvent(sk, BrainmapRealWorld.sketchApp));
			
			if(!e.editShapeOnly){
				GraphKnowledge gk = this.generateNode(e.s, sk.creationTime());
				postEvent(new KnowledgeDisseminationEvent(gk, BrainmapRealWorld.graphApp));
			}	
		} 
		
		else if (e.s instanceof ClosedShape) {
			postEvent(new KnowledgeDisseminationEvent(sk, BrainmapRealWorld.sketchApp));
			
			if(!e.editShapeOnly){
				GraphKnowledge gk = this.generateNode(e.s, sk.creationTime());
				postEvent(new KnowledgeDisseminationEvent(gk, BrainmapRealWorld.graphApp));
			}
		} 
		
		else if (e.s instanceof Line) {	
			if(e.checkDuplication && containsEdge(e.n1, e.n2)){
				postEvent(new ErrorMsgEvent("Duplicated Edge"));
				return;
			}
			
			else if(!e.editShapeOnly && null == e.n1 && null == e.n2){
				postEvent(new ErrorMsgEvent("Dangling Edge"));
				return;
			}
			
			else{
				postEvent(new KnowledgeDisseminationEvent(sk, BrainmapRealWorld.sketchApp));	
				if(!e.editShapeOnly){
					GraphKnowledge gk = this.generateEdge((Line)e.s, e.n1, e.n2, sk.creationTime());
					postEvent(new KnowledgeDisseminationEvent(gk, BrainmapRealWorld.graphApp));
				}
			}
		}
	}

	private GraphKnowledge generateEdge(Line l, String n1, String n2, double creationTime) {
		return new GraphKnowledge(l.id, TYPE.EDGE, n1, n2, creationTime);
	}

	private GraphKnowledge generateNode(Shape s, double creationTime){
		return new GraphKnowledge(s.id, TYPE.NODE, null, null, creationTime);
	}
	
	private boolean containsEdge(String n1, String n2){
		return BrainmapRealWorld.graphApp.containsEdge(n1, n2);
	}
	
	private void initQueues() {
		shapeKnowledgeWaitingQueue = new WaitingQueue("ShapeKnowledge", true);
		graphKnowledgeWaitingQueue = new WaitingQueue("GraphKnowledge", true);
		shapeKnowledgeWaitingQueue.setWaitingQueue(graphKnowledgeWaitingQueue);
		graphKnowledgeWaitingQueue.setWaitingQueue(shapeKnowledgeWaitingQueue);
		shapeKnowledgeWaitingQueue.setWaitingQueueListener(this);
		graphKnowledgeWaitingQueue.setWaitingQueueListener(this);
	}
	
	private void initAppKB(){
		appKB = new AppKnowledgeBase(){
			public void handleKnowledge(ApplicationKnowledge k){
				if(k instanceof SessionKnowledge)
					handleSessionKnowledge((SessionKnowledge)k);
				
				else if(k instanceof SessionBeaconKnowledge)
					handleSessionBeaconKnowledge((SessionBeaconKnowledge) k);	
			}
		};
	}

	@Override
	public void onMet(Waitable w1, Waitable w2) {
		ShapeKnowledge s = null;
		GraphKnowledge g = null;
		
		if (w1 instanceof ShapeKnowledge && w2 instanceof GraphKnowledge) {
			s = (ShapeKnowledge) w1;
			g = (GraphKnowledge) w2;
		} 
		
		else if (w2 instanceof ShapeKnowledge && w1 instanceof GraphKnowledge) {
			s = (ShapeKnowledge) w2;
			g = (GraphKnowledge) w1;
		} 
		
		else {
			System.err.println("BrainmapApp::onMet:Invalid type " + w1 + ", " + w2);
			return;
		}

		postEvent(new KnowledgeAcceptanceEvent(s, BrainmapRealWorld.sketchApp));
		postEvent(new KnowledgeAcceptanceEvent(g, BrainmapRealWorld.graphApp));
	}
	
	@Override
	public void onHistory(Waitable w){
		if(w instanceof ShapeKnowledge)
			postEvent(new KnowledgeAcceptanceEvent((ShapeKnowledge)w, BrainmapRealWorld.sketchApp));
		
		else if(w instanceof GraphKnowledge)
			postEvent(new KnowledgeAcceptanceEvent((GraphKnowledge)w, BrainmapRealWorld.graphApp));
	}
}
