/*

   Copyright 2012-2013 SRI International

   Licensed under the Apache License, Version 2.0 (the "License"); you may
   not use this file except in compliance with the License. You may obtain
   a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

   The development was supported by National Science Foundation Grant 0932397:
   A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.app.brainmap;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;
import java.util.Iterator;

import cf.api.*;
import cf.app.brainmap.events.RestoreEvent;
import cf.real.CyberEngineIdList;
import cf.util.Dist;
import cf.util.Log;


/*
	  CDApp implements collaborative dissemination to decrease the latency by 
	fragmenting knowledge into smaller blocks based on user-defined block size 
	(e.g., 60KBytes). Fragmented blocks are disseminated as BlockKnowledge in a 
	random order to reduce the probability of receiving the same sequence of 
	BlockKnowledge from multiple neighbors.
*/
public class CDApp extends CyberApp {
	
	// Block Table Default Row Capacity
	private static final int DEFAULT_TABLE_CAPACITY = 15;
	
	// Block Table 
	private Hashtable<String, ArrayList<BlockKnowledge>> blocktable;
	
	private AppKnowledgeBase appKB;

	public CDApp(CyberNode node) {
		super(node);
		blocktable = new Hashtable<String, ArrayList<BlockKnowledge>>(DEFAULT_TABLE_CAPACITY);
		initAppKB();
	}
	
	private void initAppKB(){
		appKB = new AppKnowledgeBase(){
			public void handleKnowledge(ApplicationKnowledge k){
				if (k instanceof BlockKnowledge){
					handleBlockKnowledge((BlockKnowledge) k);
				}
			}
		};
	}
	
	@Override
	public String name() {
		return super.name() + "::CDApp";
	}

	@Override
	public void handleEvent(ApplicationEvent e) {
		if(e instanceof CDRequestEvent)
			this.disseminate(((CDRequestEvent) e).k, ((CDRequestEvent) e).id, ((CDRequestEvent) e).blockSize);
	}
	
	@Override
	public void handleKnowledge(ApplicationKnowledge k) {
		if(k instanceof BlockKnowledge)
			appKB.updateKnowledge(k);
	}
	
	public void disseminate(ApplicationKnowledge k, String id, final int blocksize) {
		id = k.getClass().getName()+id;
		int size = knowledgeSize(k);
		
		if(-1 == size){
			Log.stderr.println("CDApp::disseminate:Failed to calculate Knowledge size..");
			return;
		}
		
		if( size <= blocksize)
			postKnowledge(k);
		
		else{
			ArrayList<BlockKnowledge> bks = BlockKnowledge.blockKnowledgeFactory(k, id, blocksize);
		
			// Post BlockKnowlege in random sequence
			Collections.shuffle(bks);
			for(int index=0; index<bks.size(); index++)
				postKnowledge(bks.get(index));
		}
	}

	// If all blocks are received, return true. Otherwise return false
	private void handleBlockKnowledge(BlockKnowledge cdk) {
		// Add to table
		
		ArrayList<BlockKnowledge> blocks = addBlockKnowledgeToTable(cdk);
		if(null == blocks)
			return;
		
		ApplicationKnowledge k = restore(blocks, cdk.totalNum);
		
		// Remove the blocks from the table
		blocks = null;
		removeFromTables(cdk.id);
		
		if (null == k)
			Log.stderr.println("CDisseminationApp::handleBlockKnowledge:Failed to restore the knowledge");
		
		else
			postEvent(new RestoreEvent(k));
	}
	
	// If the all blocks are gathered, return the blocks. Otherwise return null.
	//private BlockKnowledge[] addBlockKnowledgeToTable(BlockKnowledge bk){
	private ArrayList<BlockKnowledge> addBlockKnowledgeToTable(BlockKnowledge bk){
		//BlockKnowledge[] blocks = blocktable.get(cdk.id);
		ArrayList<BlockKnowledge> blocks = blocktable.get(bk.id);
		
		// Add new entry to table
		if (null == blocks) {
			if (bk.totalNum > 1) {
				blocks = new ArrayList<BlockKnowledge>(bk.totalNum);
				
				blocks.add(bk);
				blocktable.put(bk.id, blocks);
				
				return null;
			}
			
			// Don't need to create an entry
			else if (1 == bk.totalNum)
				blocks = new ArrayList<BlockKnowledge>(1);
			
			else {
				Log.stderr.println("CDApp::handleKnowledge:Invalid Total Number of Blocks");
				return null;
			}
		}
		
		blocks.add(bk);
		
		if (bk.totalNum != blocks.size())
			return null;
		return blocks;
	}
	
	private void removeFromTables(String id){
		blocktable.remove(id);
	}
	
	// If there is an exception, returns -1. Otherwise returns knowledge size
	private int knowledgeSize(ApplicationKnowledge k){
		try {
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			ObjectOutputStream os = new ObjectOutputStream(bos);
			os.writeObject(k);
			os.flush();
			
			int size =  bos.toByteArray().length;
			
			os.close();
			bos.close();
			
			return size;
		}
		
		catch (IOException e) {
			e.printStackTrace();
			return -1;
		}
	}

	// If there is an error, return null
	//private ApplicationKnowledge restore(BlockKnowledge[] cdks,	int totalNum) {
	private ApplicationKnowledge restore(ArrayList<BlockKnowledge> bks, int totalNum) {
		if (bks.size() < 1 || bks.size() != totalNum)
			return null;
		
		if (bks.get(0).totalNum != totalNum)
			return null;
		
		for (int index = 1; index < totalNum; index++) {
			if (!bks.get(0).related(bks.get(index)))
				return null;
		}

		// Sorting
		if (totalNum > 1)
			Collections.sort(bks);

		// Restore data
		byte[] data = null;
		try {
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			for (int index = 0; index < totalNum; index++)
				bos.write(bks.get(index).block);
			bos.flush();
			data = bos.toByteArray();
			bos.close();
		} catch (Exception e) {
			Log.stderr.println("CDApp::restore:Failed to restore the data.");
			e.printStackTrace();
			return null;
		}

		// Restore meta data
		ApplicationKnowledgeAttributes attr = bks.get(0).getApplicationKnowledgeAttributes();
		
		try {
			for (int index = 1; index < totalNum; index++)
				if (!attr.and(bks.get(index)))
					throw new Exception("Failed to restore ApplicationKnowledge metadata");
		} 
		
		catch (Exception e) {
			e.printStackTrace();
			Log.stderr.println("CDApp::restore:Failed to restore the meta data.");
			return null;
		}

		// Return
		return restore(data, attr);
	}
	
	private ApplicationKnowledge restore(byte[] data, ApplicationKnowledgeAttributes attr) {
		try {
			ObjectInputStream ois = new ObjectInputStream(new ByteArrayInputStream(data));
			ApplicationKnowledge k = (ApplicationKnowledge) ois.readObject();
			
			if (!attr.setMetadataToKnowledge(k)) {
				Log.stderr.println("CDApp::restore:Attributes are not matched with data");
				ois.close();
				return null;
			}
			
			ois.close();
			
			return k;
		} 
		catch (Exception e) {
			e.printStackTrace();
			Log.stderr.println("CDApp::restore:Failed to generate application knowledge from bytes");
		}
		return null;
	}
}

//class ApplicationKnowledgeMetaData{
class ApplicationKnowledgeAttributes{
	
	// If the knowledges are same, these are also same.
	public String creator;
	public double creationTime;
	public double activationTime;
	public double expirationTime;
	
	// It could be different even if the knowledges are same with each other.
	public CyberEngineIdList aware_engine;
    
    public ApplicationKnowledgeAttributes(ApplicationKnowledge k){
    	creator = k.creator();
        creationTime = k.creationTime();
        activationTime = k.activationTime();
        expirationTime = k.expirationTime();
        aware_engine = k.aware_engine;    
    }
    
    public ApplicationKnowledgeAttributes(ApplicationKnowledgeAttributes attr){
    	creator = attr.creator;
        creationTime = attr.creationTime;
        activationTime = attr.activationTime;
        expirationTime = attr.expirationTime;
        aware_engine = attr.aware_engine;    
    }
    
    public static CyberEngineIdList and(CyberEngineIdList l1, CyberEngineIdList l2){
    	CyberEngineIdList merged_aware_engine = new CyberEngineIdList();
    	CyberEngineIdList list1;
    	CyberEngineIdList list2;
    	if(l1.size() > l2.size()){
    		list1 = l2;
    		list2 = l1;
    	}else{
    		list1 = l1;
    		list2 = l2;
    	}
		Iterator<Long> it = list1.iterator();
		while(it.hasNext()){
			Long engine_id = it.next();
			if(list2.contains(engine_id))
				merged_aware_engine.add(engine_id);
		}
		return merged_aware_engine;
    }
    
    public void and(CyberEngineIdList p_aware_engine){
    	aware_engine = and(aware_engine, p_aware_engine);
    }
    
    public boolean and(ApplicationKnowledge k){
    	return and(new ApplicationKnowledgeAttributes(k));
    }
    
    // If meta datas are related, returns conjunction. Otherwise returns null.
    public boolean and(ApplicationKnowledgeAttributes attr){
    	if(!related(attr)){
    		Log.stderr.println("ApplicationKnowledgeMetaData::and:MetaData is not related "+toString());
    		return false;
    	}
    	
    	and(attr.aware_engine);
		return true;
    }
    
    public boolean setMetadataToKnowledge(ApplicationKnowledge k){
    	if(related(k))
    		k.aware_engine = this.aware_engine;
    	else
    		return false;
    	return true;
    }
        
    public boolean related(ApplicationKnowledgeAttributes attr){
    	if(creator.equals(attr.creator) && creationTime == attr.creationTime
    		&& activationTime == attr.activationTime && expirationTime == attr.expirationTime)
    		return true;
    	return false;
    }
    
    public boolean related(ApplicationKnowledge k){
    	if(k.creator().equals(creator) && k.creationTime()==creationTime 
    		&& k.activationTime()== activationTime && k.expirationTime()==expirationTime)
    		return true;
    	return false;
    }
    
    public String toString(){
    	return creator+","+creationTime+","+activationTime+","+expirationTime+","+aware_engine;
    }
}

class BlockKnowledge extends EditableKnowledge implements Comparable<BlockKnowledge>{
	
	//Data from ApplicationKnowledge class is about 0.9k
	private static final long serialVersionUID = 1L;
	
	// Total block number of the ApplicationKnowledge splitted
	public int totalNum;
	
	// Block Number
	public int blockNum;
	
	// Block
	public byte[] block;
		
	// Set the meta data as same with Application Knowledge
	public BlockKnowledge(String id, int blockNum, int totalNum, byte[] block, ApplicationKnowledgeAttributes attr){
		super(id, attr.creator, attr.creationTime, attr.expirationTime);
		this.block = block;
		this.blockNum = blockNum;
		this.totalNum = totalNum;
        aware_engine = attr.aware_engine;
	}
	
	@Override
	public boolean subsumes(ApplicationKnowledge k){
		return false;
	}
	
	@Override
	public boolean related(ApplicationKnowledge k){
		if(super.related(k) && k instanceof BlockKnowledge){
			if(((BlockKnowledge) k).totalNum == totalNum)
				return true;
		}
		return false;
	}
	
	@Override
	public boolean replaces(ApplicationKnowledge k){
		if(super.replaces(k) && ((BlockKnowledge) k).blockNum == blockNum)
			return true;
		return false;
	}
	
	@Override
	public int compareTo(BlockKnowledge bk){
		if(blockNum < bk.blockNum)
			return -1;
		else if(blockNum == bk.blockNum)
			return 0;
		else
			return 1;
	}
	
	@Override
	public String toString(){
		return "BlockKnowledge::"+id+"["+blockNum+"] Size("+knowledgeSize(this)+")";
	}
	
	public ApplicationKnowledgeAttributes getApplicationKnowledgeAttributes(){
		return new ApplicationKnowledgeAttributes(this);
	}
	
	public static ArrayList<BlockKnowledge> blockKnowledgeFactory(ApplicationKnowledge k, String id, final int blocksize){
		
		// Generate blocks
		int totalNum;
		byte[] data = knowledge2byte(k);
		
		if(data.length % blocksize > 0)
			totalNum = data.length/blocksize+1;
		else
			totalNum = data.length/blocksize;
		
		
		byte[][] blocks = new byte[totalNum][];
		for(int i=0; i<totalNum-1; i++)
			blocks[i] = subBytes(data, i*blocksize, blocksize);
		blocks[totalNum-1] = subBytes(data, (totalNum-1)*blocksize, data.length - blocksize*(totalNum-1)); //last block
		
		// Generate BlockKnowledge
		ArrayList<BlockKnowledge> bks = new ArrayList<BlockKnowledge>();
		ApplicationKnowledgeAttributes attr = new ApplicationKnowledgeAttributes(k);
		
		for(int index=0; index<totalNum; index++)
			bks.add(new BlockKnowledge(id, index, totalNum, blocks[index], attr));
		
		return bks;
	}
	
	private static byte[] subBytes(byte[] src, int offset, int len){
		try{
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			bos.write(src, offset, len);
			bos.flush();
			byte[] result =  bos.toByteArray();
			bos.close();
			return result;
		}catch(Exception e){
			e.printStackTrace();
		}
		
		return null;
	}
	
	private static byte[] knowledge2byte(ApplicationKnowledge k){
		try {
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			ObjectOutputStream os = new ObjectOutputStream(bos);
			os.writeObject(k);
			os.flush();
			byte[] data = bos.toByteArray();
			os.close();
			bos.close();
			return data;
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		return null;
	}
	
	private int knowledgeSize(ApplicationKnowledge k){
	  try 
	    {
	      ByteArrayOutputStream b_out = new ByteArrayOutputStream();
	      ObjectOutputStream o_out = new ObjectOutputStream(b_out);
	      o_out.writeObject(k);
	      o_out.close();
	      return b_out.toByteArray().length;
	    }
	    catch(Exception e)
	    {
	    	Log.stderr.println("KnowledeManager.kb2byte - " + e);
	    	e.printStackTrace();
	    	return -1;
	    } 
	}
}

class CDRequestEvent extends ApplicationEvent{
	public ApplicationKnowledge k;
	public String id;
	public int blockSize;
	
	public CDRequestEvent(ApplicationKnowledge k, String id, int blockSize){
		super(BrainmapRealWorld.cdApp, 0.0);
		this.k = k;
		this.id = id;
		this.blockSize = blockSize;
	}
}
