/*

   Copyright 2012-2013 SRI International

   Licensed under the Apache License, Version 2.0 (the "License"); you may
   not use this file except in compliance with the License. You may obtain
   a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

   The development was supported by National Science Foundation Grant 0932397:
   A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.app.brainmap;

import java.util.Enumeration;
import java.util.Hashtable;

import cf.api.*;


/*
	  When a user deletes a node or edge in a map, DeleteKnowledge dk
	is generated to discard EditableKnowlege ek that is associated with the
	corresponding node or edge. In particular, dk contains a delete table dt
	with entries of ek to be deleted. This replacement is specified by partial
	ordering rule
	 
	O2 : ek(..., ctek, ..., uuid_ek ) \prec dk(..., dt_dk, ...)
	     if uuid_ek = dti_dk.uuid and ct_ek ≤ dti_dk.ct,
	     where dti_dk.uuid and dti_dk.ct represent uuid and ct of i-th entry
	     in dt_dk , respectively.
	 
	  The above rule is of importance when a user leaves the network and rejoins
	later. Since any ek that is subject to be replaced by dk will be absorbed by
	the knowledge dissemination, the synchronization issue (i.e., deleted part
	of map is disseminated from disconnected users) can be resolved by this
	mechanism. The following partial order rule is used to eliminate redundant dk.
	 
	O3 : dk(..., dt_dk, ...) \prec dk(..., dt_dk, ...) if dt_dk ⊂ dt_dk .
	 
	  In the case of incomparable dks (i.e., dt_dk \not ⊂ dt_dk and
	dt_dk \not ⊃ dt_dk ), EditApp posts a new DeleteKnowledge
	dk(..., dt_dk, ...) whose dt_dk is dt_dk1 ∪ ... ∪ dt_dkn to replace
	dk1, ..., dkn, where n is the number of local dk, using O3.
*/
public class DeleteKnowledge extends ApplicationKnowledge{
	private static final long serialVersionUID = 1L;
	private DeletingTable t;
	public DeleteKnowledge(CyberApplication app){
		super(app);
		t = new DeletingTable();
	}
	
	public DeleteKnowledge(DeletingTable t, CyberApplication app){
		super(app);
		this.t = t;
	}
	
	public DeleteKnowledge(EditableKnowledge e, CyberApplication app){
		this(app);
		t.addEntry(e);
	}
	
	public DeletingTable getDeletingTable(){
		return t;
	}
	
	public void add(EditableKnowledge e){
		t.addEntry(e);
	}
	
	@Override
	public String toString(){
		return "Delete Knowledge\n  "+t.toString();
	}
	
	@Override
	public boolean related(ApplicationKnowledge k){
		if(k instanceof EditableKnowledge || k instanceof DeleteKnowledge)
			return true;
		return false;
    }

	@Override
    public boolean subsumes(ApplicationKnowledge k){
		// Currently, sumsumes() doen't remove subsumed knowledge. 
        // So below moves to replaces().
    	/*if(k instanceof DeleteKnowledge){
        	DeletingTable tk = ((DeleteKnowledge)k).t;
        	if(t.contains(tk))
        		return true;
        }*/
        return false;
    }

	@Override
    public boolean replaces(ApplicationKnowledge k){
        if(k instanceof EditableKnowledge){
        	if(t.contains((EditableKnowledge)k)){
        		return true;
        	}
        }
        
        // Currently, sumsumes() doen't remove subsumed knowledge. 
        // So below is moved from subsumes().
        else if(k instanceof DeleteKnowledge){
        	DeletingTable tk = ((DeleteKnowledge)k).t;
        	if(t.contains(tk))
        		return true;
        }
        return false;
    }
	
	@Override
    public void onReplaced(){
    	//System.out.println("DeleteKnowledge "+ this+" is replaced.");
    }
}

class DeletingTable extends Hashtable<String,Double>{
	private static final long serialVersionUID = 1L;

	public boolean contains(EditableKnowledge ek){
		Double creationTime = get(key(ek));
		if(null == creationTime)
			return false;
		if(ek.creationTime() > creationTime.doubleValue()){
			return false;
		}
		return true;
	}
	
	public DeletingTable merge(DeletingTable t){
		DeletingTable newTable = new DeletingTable();
		
		// copy
		Enumeration<String> e = keys();
		while(e.hasMoreElements()){
			String key = e.nextElement();
			Double value = get(key);
			if(null != value)
				newTable.put(key, value);
		}
		
		// merge
		e = t.keys();
		while(e.hasMoreElements()){
			String key = e.nextElement();
			Double existingCreationTime = newTable.get(key);
			Double newCreationTime = t.get(key);
			if(null == existingCreationTime)
				newTable.put(key, newCreationTime);
			else if(existingCreationTime.doubleValue() < newCreationTime.doubleValue()){
				newTable.remove(key);
				newTable.put(key, newCreationTime);
			}
		}
		return newTable;
	}
	
	public void addEntry(EditableKnowledge ek){
		String key = key(ek);
		put(key, ek.creationTime());
	}
	
	private String key(EditableKnowledge ek){
		return ek.id+":"+ ek.getClass().getName();
	}
	
	public boolean contains(DeletingTable t){
		for(String k : t.keySet()){
			Double myCreationTime = get(k);
			if(null == myCreationTime)
				return false;
			else{
				Double creationTime = t.get(k);
				if(creationTime.doubleValue() != myCreationTime)
					return false;
			}
		}
		return true;
	}
	
	@Override
	public boolean equals(Object o){
		if(o instanceof DeletingTable){
			DeletingTable t = (DeletingTable)o;
			if(!t.contains(this))
				return false;
			if(!contains(t))
				return false;
			return true;
		}
		return false;
	}
	
	@Override
	public String toString(){
		String str = "  ";
		for(String key : keySet())
			str += key+"::"+get(key)+"\n";
		return str;
	}	
}
