/*

   Copyright 2012-2013 SRI International

   Licensed under the Apache License, Version 2.0 (the "License"); you may
   not use this file except in compliance with the License. You may obtain
   a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

   The development was supported by National Science Foundation Grant 0932397:
   A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.app.brainmap;

import java.util.ArrayList;
import java.util.Iterator;


public class ObjectQuadtree {
	
	// Threshold to figure out object selection
	private static final double MIN_SELECT_DISTANCE = 25.0;
	
	// Number of points to store in a node. It's same with the number of duplicated points to save
	private final int NODE_CAPACITY;	
	
	// Tree root node
	private Node root;
	
	// Space covered by tree.
	private double width;
	private double height;
	private Point2df lt;
	private Point2df rb;
	
	public ObjectQuadtree(double width, double height, int nodeCapacity){
		NODE_CAPACITY = nodeCapacity;
		
		this.width = width;
		this.height = height;
		lt = new Point2df((float)(-width/2.0), (float)(height/2.0));
		rb = new Point2df((float)(width/2.0), (float)(-height/2.0));
		
		root = new Node(0, new Point2df(0.0f, 0.0f), null);
	}
	
	synchronized public void insert(Object o, Point2df p){
		if(p.x < lt.x || rb.x<p.x || p.y < rb.y || lt.y < p.y)
			scale();
		root.insertPoint(o, p);
	}
	
	synchronized public void delete(Object o){
		root.delete(o);
	}
	
	public Object select(Point2df p, boolean nearest){
		return root.select(p, null, nearest);
	}
	
	public Object select(Point2df p, Class c){
		return root.select(p, c, true);
	}
	
	public ArrayList<Object> getObjects(){
		return root.getObjects();
	}

	public Iterator<Node> getLeaves(){
		return new Leaves();
	}
	
	public void print(){
		System.out.println(root.toString());
	}

	synchronized private void scale(){
		this.width = width*2.0;
		this.height = height*2.0;
		lt = new Point2df((float)(-width/2.0), (float)(height/2.0));
		rb = new Point2df((float)(width/2.0), (float)(-height/2.0));
		
		Node newroot = new Node(0, new Point2df(0.0f, 0.0f), null);
		
		Iterator<Node> itor = this.getLeaves();
		while(itor.hasNext()){
			Node n = itor.next();
			for(int i=0; i<n.in; i++)
				newroot.insertPoint(n.o, n.points[i]);
		}
		
		root = newroot;
	}
	
	public class Leaves implements Iterator<Node>{
		private ArrayList<Node> nodes;
		private int out = 0;
		
		public Leaves(){
			nodes = new ArrayList<Node>();
			traverse(root);
		}
		
		private void traverse(Node n){
			if(n.isLeaf()){
				nodes.add(n);
				return;
			}
			traverse(n.children[0]);
			traverse(n.children[1]);
			traverse(n.children[2]);
			traverse(n.children[3]);
		} 
		
		@Override
		public boolean hasNext(){
			if(nodes.size() > out)
				return true;
			return false;
		}
		
		@Override
		public Node next(){
			return nodes.get(out++);
		}
		
		@Override
		public void remove(){}
	}
	
	class Node{
		int in;
		
		// Object contaning the node
		Object o;
		
		// Tree depth
		int depth;
		
		// Center point of the space covered by node
		Point2df center;
		
		Point2df[] points = null;
		Object[] objects = null;
		
		/* Children index
		 *    ^ 
		 *  0 | 1
		 * ---|--->
		 *  3 | 2
		 * */
		Node[] children;
		Node parent;
		
		public Node(int depth, Point2df center, Node parent){
			init(depth, center, parent);
		}
		
		public boolean isLeaf(){
			if(null == children)
				return true;
			return false;
		}
		
		public ArrayList<Object> getObjects(){
			ArrayList<Object> result;
			result = new ArrayList<Object>();
			
			if(null != o){
				result.add(o);
				return result;
			}
			
			if(isLeaf()){
				for(int i=0; i<NODE_CAPACITY; i++){
					if(null != objects[i] && !result.contains(objects[i]))
						result.add(objects[i]);
				}
			}else{
				result = children[0].getObjects();
				for(int i=1; i<4; i++){
					ArrayList<Object> os = children[i].getObjects();
					
					if(null == os)
						continue;
					
					for(Object oo : os){
						if(null != oo && !result.contains(oo))
							result.add(oo);
					}
				}
			}
			return result;
		}
		
		public Object select(Point2df p, Class c, boolean nearest){
			if(null != o && nearest){
				if(null == c)
					return o;
				else if(o.getClass() == c)
					return o;
			}else if(null != children){
				Object ret = null;
				if(p.x < center.x){
					if(p.y > center.y){
						// Backtracking
						for(int i=0; i<4 && null == ret; i++)	
							ret = children[(0+i)%4].select(p, c, nearest);
						return ret;
					}else{
						for(int i=0; i<4 && null == ret; i++)	
							ret = children[(3+i)%4].select(p, c, nearest);
						return ret;
					}
				}else{
					if(p.y > center.y){
						for(int i=0; i<4 && null == ret; i++)	
							ret = children[(1+i)%4].select(p, c, nearest);
						return ret;
					}else{
						for(int i=0; i<4 && null == ret; i++)	
							ret = children[(2+i)%4].select(p, c, nearest);
						return ret;
					}
				}
			}else if(isLeaf()){
				double mind = Double.MAX_VALUE;
				Object mino = null;
				for(int i=0; i<in; i++){
					double d = p.length(points[i]);
					if(mind > d){
						mind = d;
						mino = objects[i];
					}
				}
				if(nearest){
					if(null == c)
						return mino;
					else if(mino != null && mino.getClass() == c)
						return mino;
				}else if(mind < MIN_SELECT_DISTANCE){
					if(null == c)
						return mino;
					else if(mino != null && mino.getClass() == c)
						return mino;
				}
			}
			return null;
		}
		
		private void init(int depth, Point2df center, Node parent){
			init();
			if(null == points){
				points = new Point2df[NODE_CAPACITY];
				for(int i=0; i<NODE_CAPACITY; i++)
					points[i] = null;
			}
			if(null == objects){
				objects = new Object[NODE_CAPACITY];
				for(int i=0; i<NODE_CAPACITY; i++)
					objects[i] = null;
			}
			this.center = center;
			this.parent = parent;
			this.depth = depth;
		}
		
		private void init(){
			in = 0;
			o = null;
			children = null;
		}
		
		public void delete(Object o){
			if(this.o == o)
				init();
			else if(isLeaf()){
				// Delete Points corresponding to o
				Object oo = null;
				Object representativeO = null;
				for(int i=0; i<in; i++){
					Object oi = objects[i];
					if(o == oi)
						deletePoint(i);
					else{
						if(null == oo){
							oo = oi;
							representativeO = oi;
						}
						else if(oi != oo)
							representativeO = null;							
					}
				}
				this.o = representativeO;
			}else{
				children[0].delete(o);
				children[1].delete(o);
				children[2].delete(o);
				children[3].delete(o);
			}
		}
		
		public void insertPoint(Object o, Point2df p){
			if(null != children){
				if(o != this.o)
					this.o = null;
				
				insertToChild(o, p);
			}
			else{
				for(int i=0; i<in; i++){
					if(p.equals(points[i]))	// discard duplicated point
						return;
				}
				
				if(in >= NODE_CAPACITY){
					if(o != this.o)
						this.o = null;
					
					if(isLeaf())
						split();
					
					for(int i=0; i<NODE_CAPACITY; i++)
						insertToChild(objects[i], points[i]);
					
					insertToChild(o, p);
					in = 0;
				}
				else{
					if(0==in && null==this.o)
						this.o = o;
					
					else if(null!=this.o && o!=this.o)
						this.o = null;
					
					objects[in] = o;
					points[in++] = p;
				}
			}
		}
		
		private void insertToChild(Object o, Point2df p){
			if(isLeaf()){
				try {
					throw new Exception("Can't insert point to child: It's leaf node");
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			
			if(p.x < center.x){
				if(p.y > center.y)
					children[0].insertPoint(o, p);
				else
					children[3].insertPoint(o, p);
			}else{
				if(p.y > center.y)
					children[1].insertPoint(o, p);
				else
					children[2].insertPoint(o, p);
			}
		}
		
		private void split(){
			float dx = (float)(width/(Math.pow(2.0,(double)(depth+2))));
			float dy = (float)(height/(Math.pow(2.0,(double)(depth+2))));
			int depth = this.depth+1;
			
			children = new Node[4];
			children[0] = new Node(depth, new Point2df(center.x-dx, center.y+dy), this);
			children[1] = new Node(depth, new Point2df(center.x+dx, center.y+dy), this);
			children[2] = new Node(depth, new Point2df(center.x+dx, center.y-dy), this);
			children[3] = new Node(depth, new Point2df(center.x-dx, center.y-dy), this);
			
		}
		
		private void deletePoint(int index){
			for(int i=index; i<in-1; i++){
				points[i] = points[i+1];
				objects[i] = objects[i+1];
			}
			in--;
		}
		
		@Override
		public String toString(){
			String str = "";
			for(int i=0; i<depth; i++)
				str += "  ";
			
			float dx = (float)(width/(Math.pow(2.0,(double)(depth+1))));
			float dy = (float)(height/(Math.pow(2.0,(double)(depth+1))));
			
			Point2df lt = new Point2df(center.x - dx, center.y + dy);
			Point2df rb = new Point2df(center.x + dx, center.y - dy);
			
			str += "["+depth+"] c"+center+" in("+in+") x("+lt.x+","+rb.x+") y("+rb.y+","+lt.y+") o("+o+")\n";
			
			if(null != children){
				for(int i=0; i<4; i++)
					str += children[i].toString();
			}
			
			return str;
		}
	}
}
