/*

   Copyright 2012-2013 SRI International

   Licensed under the Apache License, Version 2.0 (the "License"); you may
   not use this file except in compliance with the License. You may obtain
   a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

   The development was supported by National Science Foundation Grant 0932397:
   A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.app.brainmap;

import java.util.ArrayList;

import cf.app.brainmap.events.PopupMenuEvent;

public class PopupMenu{
	// Menu Parameters
	public static final String MENU_SEPARATOR=",";
	public static final int MENU_NUM = PopupMenuEvent.menus.length;
	public static final int MARGIN = 10;
	private final int WIDTH;
	private final int HEIGHT;
	private final int ONE_MENU_HEIGHT;
	private final Color color = Color.ORANGE;
	
	// Menu arguments
	private final Point2df center;
	private final Point2df screenLt;
	private final Point2df lt;
	
	// Palette Parameters
	private final double COLOR_R = 10;
	private final double COLOR_ANGLE_MARGIN = 3.0;
	
	// Members for Palette
	private double r;
	private double startAngle;
	private double endAngle;
	private ArrayList<Color> allColors;
	private boolean isPaletteEnabled = false;
	
	// Screen info
	private int screenWidth;
	private int screenHeight;
	
	public PopupMenu(Point2df menu_lt_on_map, Point2df screen_lt_on_map, int width, int height, int screenWidth, int screenHeight){
		WIDTH = width;
		HEIGHT = height;
		ONE_MENU_HEIGHT = height/MENU_NUM;
		
		this.lt = menu_lt_on_map;
		screenLt = screen_lt_on_map;
		
		this.screenWidth = screenWidth;
		this.screenHeight = screenHeight;
		
		if(lt.x + WIDTH > screen_lt_on_map.x + screenWidth)
			lt.x = lt.x - WIDTH;
		if(lt.y-HEIGHT < screen_lt_on_map.y - screenHeight)
			lt.y = lt.y + HEIGHT;
		
		center = new Point2df((float)WIDTH/2.0F + lt.x, lt.y - (float)HEIGHT/2.0F);
		
		initPalette();
	}
	
	public void paint(GuiAdapter gui){
		// Draw menu
		gui.drawFilledRect(lt, WIDTH, HEIGHT, color);
		Point2df[] line = new Point2df[2];
		line[0] = new Point2df(lt.x, lt.y-ONE_MENU_HEIGHT);
		line[1] = new Point2df(lt.x+WIDTH, lt.y-ONE_MENU_HEIGHT);
		gui.drawLine(line, 2, Color.WHITE);
		
		// Draw menu items
		for(int i=0; i<MENU_NUM; i++)
			gui.drawString(PopupMenuEvent.menus[i], ALIGN.LEFT_TOP,	new Point2df(MARGIN+lt.x, lt.y-MARGIN-ONE_MENU_HEIGHT*i), Color.WHITE);
		
		// Draw palette
		if(isPaletteEnabled){
			int i=0;
			int startIndex = 0;
			double r = this.r;
			double end;
			if(startAngle > endAngle)	
				end = endAngle + 360.0;
			else
				end = endAngle;
			
			while(i < allColors.size()){
				double alpha = 180.0 - 2.0*Math.acos(COLOR_R/r/2.0)*180.0/Math.PI;
				double angle = (startAngle + (double)(i-startIndex) * (alpha*2.0+COLOR_ANGLE_MARGIN))%360.0;
				
				if(angle + alpha > end){
					startIndex = i;
					r = r + COLOR_R*2.0+COLOR_ANGLE_MARGIN;
					double[] angles = getPaletteAngleBound(r);
					startAngle = angles[0];
					endAngle = angles[1];
				}else{
					float x = (float)(r*Math.cos(angle*Math.PI/180.0)+center.x);
					float y = (float)(r*Math.sin(angle*Math.PI/180.0)+center.y);
					gui.drawFilledCircle(new Point2df(x, y), (float)COLOR_R, allColors.get(i));
					
					i++;
				}
			}
		}
	}
	
	public String selectMenu(Point2df point_on_map){
		// Select Menu Item
		if((int)lt.x < point_on_map.x && point_on_map.x < (int)lt.x+WIDTH 
				&& (int)lt.y > point_on_map.y && point_on_map.y > (int)lt.y-HEIGHT){
			for(int i=0; i<MENU_NUM; i++){
				if((int)lt.y > point_on_map.y-ONE_MENU_HEIGHT*i && point_on_map.y>(int)lt.y-ONE_MENU_HEIGHT*(i+1)){
					return PopupMenuEvent.menus[i]; 
				}
			}
		}
		
		// Select Color on palette
		else if(isPaletteEnabled){
			int i=0;
			int startIndex = 0;
			double r = this.r;
			double end;
			if(startAngle > endAngle)
				end = endAngle + 360.0;
			else
				end = endAngle;
			
			double[] angles = getPaletteAngleBound(r);
			startAngle = angles[0];
			endAngle = angles[1];
			
			while(i < allColors.size()){
				double alpha = 180.0 - 2.0*Math.acos(COLOR_R/r/2.0)*180.0/Math.PI;
				double angle = (startAngle + (double)(i-startIndex) * (alpha*2.0+COLOR_ANGLE_MARGIN))%360.0;
				
				if(angle + alpha > end){
					startIndex = i;
					r = r + COLOR_R*2.0+COLOR_ANGLE_MARGIN;
					angles = getPaletteAngleBound(r);
					startAngle = angles[0];
					endAngle = angles[1];
				}else{
					double x = (r*Math.cos(angle*Math.PI/180.0)+center.x);
					double y = (r*Math.sin(angle*Math.PI/180.0)+center.y);
					double dx_sq = Math.pow((double)point_on_map.x-x, 2.0);
					double dy_sq = Math.pow((double)point_on_map.y-y, 2.0);
					double r_sq = Math.pow(COLOR_R, 2.0);
					
					if(dx_sq + dy_sq <= r_sq)
						return "Edit Color"+MENU_SEPARATOR+allColors.get(i).toString();
					i++;
				}
			}
		}
		return null;
	}
	
	public void showPalette(){
		isPaletteEnabled = true;
	}
	
	private void initPalette(){
		r = Math.sqrt(Math.pow(WIDTH, 2.0)+Math.pow(HEIGHT, 2.0))/2.0+COLOR_R*1.5;
		allColors = new ArrayList<Color>();
		allColors.add(Color.BLACK);
		allColors.add(Color.DARK_GRAY);
		allColors.add(Color.GREEN);
		allColors.add(Color.LIGHT_GRAY);
		allColors.add(Color.MAGENTA);
		allColors.add(Color.ORANGE);
		allColors.add(Color.PINK);
		allColors.add(Color.RED);
		allColors.add(Color.YELLOW);
		allColors.add(Color.BLUE);
		allColors.add(Color.GRAY);
		
		double[] angles = getPaletteAngleBound(r);
		startAngle = angles[0];
		endAngle = angles[1];
	}
	
	private double[] getPaletteAngleBound(double r){
		double a1;
		double a2;
		double startAngle, endAngle;
		
		// Set bound
		if(r+COLOR_R > (center.x-screenLt.x)){
			a1 = Math.acos((center.x-screenLt.x) / (r+COLOR_R))*180.0/Math.PI;
			if(r+COLOR_R > (screenLt.y-center.y)){	//NW
				a2 = Math.asin((screenLt.y-center.y) / (r+COLOR_R))*180.0/Math.PI;
				startAngle = 180.0+a1;
				endAngle = 90.0-a2;
			}else if(r+COLOR_R > (center.y-(screenLt.y-screenHeight))){	//SW
				a2 = Math.asin((center.y-(screenLt.y-screenHeight)) / (r+COLOR_R))*180.0/Math.PI;
				startAngle = 270.0+a2;
				endAngle = 180.0-a1;
			}else{	//W
				startAngle = 180.0+a1;
				endAngle = 180.0-a1;
			}
		}
		
		else if(r+COLOR_R > (screenLt.x+screenWidth-center.x)){
			a1 = Math.acos((screenLt.x+screenWidth-center.x) / (r+COLOR_R))*180.0/Math.PI;
			if(r+COLOR_R > (screenLt.y-center.y)){	//NE
				a2 = Math.asin((screenLt.y-center.y) / (r+COLOR_R))*180.0/Math.PI;
				startAngle = 90.0 + a2;
				endAngle = 360.0 - a1;
			}else if(r+COLOR_R > (center.y-(screenLt.y-screenHeight))){	//SE
				a2 = Math.asin((center.y-(screenLt.y-screenHeight)) / (r+COLOR_R))*180.0/Math.PI;
				startAngle = a1;
				endAngle = 270.0-a2;
			}else{	//E
				startAngle = a1;
				endAngle = 360.0-a1;
			}
		}
		
		else{
			if(r+COLOR_R > (screenLt.y-center.y)){	//N
				a2 = Math.asin((screenLt.y-center.y) / (r+COLOR_R))*180.0/Math.PI;
				startAngle = 90.0+a2;
				endAngle = 90.0-a2;
			}else if(r+COLOR_R > (center.y-(screenLt.y-screenHeight))){	//S
				a2 = Math.asin((center.y-(screenLt.y-screenHeight)) / (r+COLOR_R))*180.0/Math.PI;
				startAngle = 270.0+a2;
				endAngle = 270.0-a2;
			}else{	//Center
				startAngle = 271.0;
				endAngle = 269.0;
			}
		}
		
		// Set margin
		startAngle += COLOR_ANGLE_MARGIN*2;
		endAngle -= COLOR_ANGLE_MARGIN*2;
		
		double angleAmount = (endAngle-startAngle+360.0)%360.0;
		double alpha = 180.0 - 2.0*Math.acos(COLOR_R/r/2.0)*180.0/Math.PI;
		double needAngle = 2*(double)allColors.size()*alpha+COLOR_ANGLE_MARGIN*(double)(allColors.size()-1);
		
		if(needAngle < angleAmount)
			startAngle = startAngle+(angleAmount-needAngle)/2.0 + alpha;
		else{
			int firstLevelColorNum = allColors.size() - (int)((needAngle-angleAmount)/(20.*alpha+COLOR_ANGLE_MARGIN)) - 1;
			needAngle = 2*(double)firstLevelColorNum*alpha+COLOR_ANGLE_MARGIN*(double)(firstLevelColorNum-1);
			startAngle = startAngle+(angleAmount-needAngle)/2.0 + alpha;
		}
		
		double[] angles =  {startAngle, endAngle};
		return angles;
	}
}
