/*

   Copyright 2012-2013 SRI International

   Licensed under the Apache License, Version 2.0 (the "License"); you may
   not use this file except in compliance with the License. You may obtain
   a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

   The development was supported by National Science Foundation Grant 0932397:
   A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.app.brainmap;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

import cf.util.Dist;

	
public abstract class Shape implements Serializable, Cloneable{
	private static final long serialVersionUID = 1L;
	
	public String id;
	public Point2df[] points;
	public int length;
	public Color color;
	 
	public Point2df lt;
	public Point2df rb;
		
	public abstract void paint(GuiAdapter gui);
	 
	public Shape(String id, Color color, Point2df[] points, int len){
		this.id = id;
		this.color = color;
		this.points = points;
		this.length = len;
		 
		setBound();
	}
	
	public boolean contains(Point2df p){
		
		Point2df[] xline = new Point2df[2];
		Point2df[] yline = new Point2df[2];
		
		xline[0] = new Point2df(Float.NEGATIVE_INFINITY, p.y);
		xline[1] = new Point2df(Float.POSITIVE_INFINITY, p.y);
		
		yline[0] = new Point2df(p.x, Float.NEGATIVE_INFINITY);
		yline[1] = new Point2df(p.x, Float.POSITIVE_INFINITY);
		
		Hashtable<Integer,Point2df> xIntersections = getIntersections(xline, 2, false);
		Hashtable<Integer,Point2df> yIntersections = getIntersections(yline, 2, false);
		
		if(null==xIntersections || null==yIntersections)
			return false;
			
		if(xIntersections.size() < 2 || yIntersections.size() < 2)
			return false;
		
		Enumeration<Integer> e = xIntersections.keys();
		float closestLeft = Float.NEGATIVE_INFINITY;
		float closestRight = Float.POSITIVE_INFINITY;
		
		while(e.hasMoreElements()){
			Point2df xp = (Point2df)xIntersections.get(e.nextElement());
			
			if(xp.x <= p.x && closestLeft < xp.x)
				closestLeft = xp.x;
			else if(p.x <= xp.x && xp.x < closestRight)
				closestRight = xp.x;
		}
		
		if(Float.POSITIVE_INFINITY == closestRight)
			closestRight = closestLeft;
		if(Float.NEGATIVE_INFINITY == closestLeft)
			closestLeft = closestRight;
				
		if(!(closestLeft <= p.x && p.x <= closestRight))
				return false;
		
		e = yIntersections.keys();
		float closestUp = Float.POSITIVE_INFINITY;
		float closestDown = Float.NEGATIVE_INFINITY;
		
		while(e.hasMoreElements()){
			Point2df yp = (Point2df)yIntersections.get(e.nextElement());
			
			if(yp.y <= p.y && closestDown < yp.y)
				closestDown = yp.y;
			else if(p.y <= yp.y && yp.y < closestUp)
				closestUp = yp.y;
		}
		
		if(Float.POSITIVE_INFINITY == closestUp)
			closestUp = closestDown;
		if(Float.NEGATIVE_INFINITY == closestDown)
			closestDown = closestUp;
		
		if(!(closestDown<=p.y&&p.y<=closestUp))
				return false;
		
		return true;
	}
	
	public boolean contains(Shape s){
		for(Point2df p : s.points){
			if(!contains(p))
				return false;
		}
		return true;
	}
	 
	public boolean conflictWith(Shape s){
		for(Point2df p : points){
			if(s.contains(p))
				return true;
		}
		
		for(Point2df p : s.points){
			if(contains(p))
				return true;
		}
		
		return false;
	}
	 
	public void cut(boolean fromHead, Shape intersected){
		Hashtable<Integer,Point2df> table = getIntersections(intersected, fromHead);
		 
		if(null == table)
			return;
		
		Enumeration<Integer> e = table.keys();
		 
		if(!e.hasMoreElements())
			return;
		 
		int i = e.nextElement();
		if(!fromHead){
			while(e.hasMoreElements())
				i = e.nextElement();
		}
		
		Point2df intersection = table.get(i);
		 
		if(fromHead){	// cut 0 ~ intersection
			length = length-i;
			Point2df[] result = new Point2df[length];
			result[0] = intersection;
			
			for(int index=1; index<length; index++)
				result[index] = points[index+i];
			
			points = result;
		}
		else{	// cut intersection-1 ~ len-1
			length = i+1;
			Point2df[] result = new Point2df[length];
			
			for(int index=0; index<length-1; index++)
				result[index] = points[index];
			
			result[length-1] = intersection;
			points = result;
		}
		return;
	}
	 
	@Override
	public String toString(){
		return "Shape-"+id;
	}
	
	@Override
	public Shape clone(){
		try {
			return (Shape)super.clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
			return null;
		}
	}
	
	private Hashtable<Integer,Point2df> getIntersections(Shape s, boolean onlyOneIntersection){
		return getIntersections(s.points, s.length, onlyOneIntersection);
	}
	 
	 // If there is a intersection, return intersections. Otherwise, return null.
	private Hashtable<Integer,Point2df> getIntersections(Point2df[] ps, int len, boolean onlyOneIntersection){
		Hashtable<Integer,Point2df> intersections = new Hashtable<Integer,Point2df>();
		 
		Point2df[] s1 = points;
		Point2df[] s2 = ps;
		int len1 = length;
		int len2 = len;
		 
		for(int i=1; i<len1; i++){
			Point2df p1=s1[i-1];
			Point2df p2=s1[i];
			float minx = p1.x<p2.x? p1.x:p2.x;
			float maxx = p1.x<p2.x? p2.x:p1.x;
			float miny = p1.y<p2.y? p1.y:p2.y;
			float maxy = p1.y<p2.y? p2.y:p1.y;
			
			for(int j=1; j<len2; j++){
				Point2df p3=s2[j-1];
				Point2df p4=s2[j];
				float minx2 = p3.x<p4.x? p3.x:p4.x;
				float maxx2 = p3.x<p4.x? p4.x:p3.x;
				if((minx<=p3.x && p3.x <= maxx) || (minx<=p4.x && p4.x<=maxx) || (minx2<minx && maxx<maxx2)){
					float miny2 = p3.y<p4.y? p3.y:p4.y;
					float maxy2 = p3.y<p4.y? p4.y:p3.y;
					if((miny<=p3.y && p3.y <= maxy) || (miny<=p4.y && p4.y<=maxy) || (miny2<miny && maxy<maxy2)){
						float dx1 = p2.x-p1.x;
						float dx2 = p4.x-p3.x;
						float dy1 = p2.y-p1.y;
						float dy2 = p4.y-p3.y;
						float m1=(dx1==0)? Float.MAX_VALUE:dy1/dx1;
						float m2=(dx2==0)? Float.MAX_VALUE:dy2/dx2;
						float x1=p1.x;
						float x2=p3.x;
						float y1=p1.y;
						float y2=p3.y;
						float x = Float.NaN;
						float y = Float.NaN;
						if(0 == m1){
							if(0 == m2){
								if(y1 == y2){
									x = ((x1+dx1/2.0f)+(x2+dx2/2.0f)) / 2.0f;	//ave of ave
								y = y1;
							}								
							}else if(Float.MAX_VALUE == m2){
								x = x2;
								y = y1;
							}else{
								y = y1;
								x = (y-y2)/m2+x2;
							}
						}else if(Float.MAX_VALUE == m1){ 
							if(0 == m2){
								x = x1;
								y = y2;
							}else if(Float.MAX_VALUE == m2){
								if(x1==x2){
									x = x1;
									y = ((y1+dy1/2.0f)+(y2+dy2/2.0f)) / 2.0f;
								}
							}else{
								x = x1;
								y = m2*(x-x2)+y2;
							}
						}else{
							if(0 == m2){
								y = y2;
								x = (y-y1)/m1+x1;
							}else if(Float.MAX_VALUE == m2){
								x = x2;
								y = m1*(x-x1)+y1;
							}else{
								if(m1 == m2){	//paralell		
									if(y2 == m1*(x2-x1)+y1){
										x = ((x1+dx1/2.0f)+(x2+dx2/2.0f)) / 2.0f;	//ave of ave
										y = ((y1+dy1/2.0f)+(y2+dy2/2.0f)) / 2.0f;
									}
								}else{
									x = (m1*x1-m2*x2-y1+y2)/(m1-m2);
									//y = m1*(x-x1)+y1;
									y = (m1*m2*(x1-x2)+m1*y2-m2*y1)/(m1-m2);
								}
							}
						}
						if(Float.NaN == x || Float.NaN == y)
							continue;
						
						// intersection range
						float con_minx = minx<minx2? minx:minx2;
						float con_maxx = maxx<maxx2? maxx2:maxx;
						float con_miny = miny<miny2? miny:miny2;
						float con_maxy = maxy<maxy2? maxy2:maxy;
						
						// Add intersection
						if((con_minx<=x && x<=con_maxx) && (con_miny<=y && y<=con_maxy)){
							intersections.put(i, new Point2df(x, y));
							
							if(onlyOneIntersection)
								return intersections;
						}
					}
				}
			}
		}
		
		if(intersections.size() > 0)
			return intersections;
		
		return null;
	}
	 
	private void setBound(){
		float maxX = Float.NEGATIVE_INFINITY;
		float maxY = Float.NEGATIVE_INFINITY;
		float minX = Float.POSITIVE_INFINITY;
		float minY = Float.POSITIVE_INFINITY;
		 
		for(int i=0; i<length; i++){
			if(points[i].x > maxX)
				maxX = points[i].x;
			if(points[i].x < minX)
				minX = points[i].x;
			if(points[i].y > maxY)
				maxY = points[i].y;
			if(points[i].y < minY)
				minY = points[i].y;
		}
		 
		lt = new Point2df(minX, maxY);
		rb = new Point2df(maxX, minY);
	}
}

class Line extends Shape implements Cloneable{
	private static final long serialVersionUID = 1L;
	public static final int CIRCLE_R = 5;
	
	boolean head, tail;

	public Line(Color color, boolean head, boolean tail, Point2df[] points, int len){
		this(new Long(Dist.dist.rand().nextLong()).toString(), color, head, tail, points, len);
	}
	
	public Line(String id, Color color, boolean head, boolean tail, Point2df[] points, int len){
		 super(id, color, points, len);
		 this.head = head;
		 this.tail = tail;
	 }
	
	public Line(Line l, Color c){
		this(l.id, c, l.head, l.tail, l.points, l.length);
	}
	
	public void setHead(boolean head){
		this.head = head;
	}
	
	public void setTail(boolean tail){
		this.tail = tail;
	}
	
	
	@Override
	public void paint(GuiAdapter gui){
		gui.drawLine(points, length, color);
		
		if(head)
			gui.drawFilledCircle(points[0], CIRCLE_R, color);
		if(tail)
			gui.drawFilledCircle(points[length-1], CIRCLE_R, color);
	}
	
	@Override
	public String toString(){
		return "Line-"+id;
	}
	
	@Override
	public Line clone(){
		return (Line)super.clone();
	}
}

class ClosedShape extends Shape implements Cloneable{
	private static final long serialVersionUID = 1L;
	
	public Point2df textCenter;
	public String text = null;
		
	public Color fill;
	
	public ClosedShape(String id, Color fill, Color line, Point2df[] points, int len, String text, Point2df textCenter){
		super(id, line, points, len);
		this.fill = fill;
		this.text = text;
		this.textCenter = textCenter;
	}
	
	public ClosedShape(Color line, Point2df[] points, int len, Color fill, String text, Point2df textCenter){
		this(new Long(Dist.dist.rand().nextLong()).toString(), fill, line, points, len, text, textCenter);
	}
	
	public ClosedShape(Color fill, Color line, Point2df[] points, int len){
		this(new Long(Dist.dist.rand().nextLong()).toString(), fill, line, points, len, null, null);
	}
	
	public ClosedShape(ClosedShape cs, String text, Point2df textCenter, Color c){
		super(cs.id, c, cs.points, cs.length);
		this.fill = cs.fill;
		this.text = text;
		this.textCenter = textCenter;
	}
	
	public ClosedShape(ClosedShape cs, Color c){
		this(cs, cs.text, cs.textCenter, c);
		this.color = c;
	}
	
	public double area(){
		double dx = Math.abs(rb.x - lt.x);
		double dy = Math.abs(lt.y - rb.y);
		return dx*dy;
	}
	
	@Override
	public void paint(GuiAdapter gui){
		gui.drawFilledPolygon(points, length, color, fill);
		if(null != text)
			gui.drawString(text, ALIGN.CENTER, textCenter, color);
	}
	
	@Override
	public String toString(){
		return "ClosedShape-"+id;
	}
	
	@Override
	public ClosedShape clone(){
		return (ClosedShape)super.clone();
	}
	
}

class Text extends Shape implements Cloneable{
	private static final long serialVersionUID = 1L;
	
	public String text;
	
	 // points[0] is the center of the text
	public Text(String text, Color color, Point2df[] points){
		this(text, text, color, points);
	}
	
	public Text(String id, String text, Color color, Point2df[] points){
		 super(id, color, points, 1);
		 this.text = text;		 
	 }
	
	@Override
	public void paint(GuiAdapter gui){
		gui.drawString(text, ALIGN.CENTER, points[0], color);
	}
	
	@Override
	public String toString(){
		return "Text-"+id+"("+text+")";
	}
	
	@Override
	public Text clone(){
		return (Text)super.clone();
	}
}
