/*

   Copyright 2012-2013 SRI International

   Licensed under the Apache License, Version 2.0 (the "License"); you may
   not use this file except in compliance with the License. You may obtain
   a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

   The development was supported by National Science Foundation Grant 0932397:
   A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.app.brainmap;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Timer;
import java.util.TimerTask;

//import android.util.Log;

import cf.api.*;
import cf.app.brainmap.events.*;
import cf.common.KnowledgeBase;


/*
	  SketchApp generates geometry data and corresponding ContentKnowledge
	from user input. It handles ShapeKnowledge sk and ContentKnowledge ck to enable 
	pairing between them, and posts sk as DEvent that will be handled by 
	BrainmapApp. SketchApp also generates DeleteKnowledge dk to replace sk and ck 
	and posts GraphDeleteEvent when a local user eliminates a diagram.
 */
public class SketchApp extends CyberApp implements WaitingQueueListener{
	private static final boolean PRINT_STATUS = false;
	private static final double SESSION_LIST_UPDATE_DELAY = 0.5; //0.5s
	
	// Keys
	public static final int BS = 8;
	public static final int ENTER = 10;
	public static final int SPACE = 32;
	
	// Click Parameters
	public static final double CLICK_TOLERANCE = 10.0;
	public static final long DOUBLE_CLICK_TIME = 1000;
	public static final long LONG_CLICK_TIME = 1000;
	
	// Block size for Collaborative dissemination
	private static final int BLOCK_SIZ = 60000;
	
	// Point number increasing unit of temporary Shape, which shows user input directly
	public static final int TMP_SHAPE_SIZE_UNIT = 100;
	
	// Minimum area of ClosedShape
	public static final double MIN_CLOSED_SHAPE_AREA = 1000.0;
	
	private Session session = null;
	
	// GUI adapter
	private GuiAdapter gui;
	
	// File Manager
	private FileAdapter fa;
	
	// User Input State
	private boolean isDrawing = false;
	private boolean isMoving = false;
	private Point2df movingStart = null;
	private boolean isScaling = false;
	private boolean isTexting = false;
	
	// Temp Shape for showing user input directly
	private String tmpText="";
	private int tmpPointsPointer = 0;
	private Point2df[] tmpPoints = null;
	
	// Click condition
	Timer pressTimer;
	private long lastClickTime = 0;
	private boolean longPress = false;
	private Point2df lastPressPoint = null;
	private Point2df lastLongPressPoint = null;
	private Point2df lastDoubleClickPoint = null;
	
	// Queues for ShapeKnowledge-ContentKnowledge pair
	private WaitingQueue shapeWaitingQueue;
	private WaitingQueue contentWaitingQueue;
	
	// Quad tree Parameters
	private static final int QUDTREE_WIDTH = 1500;
	private static final int QUADTREE_HEIGHT = 1500;
	private static final int QUADTREE_NODE_CAPACITY = 4;
	
	// Quad tree to store node (ClosedShape, FileContentShape) only
	private ObjectQuadtree nodeQuadtree;
	
	// Ouad tree to store edge (Line) only
	private ObjectQuadtree edgeQuadtree;
	
	// Selected Shape Knowledge
	private ShapeKnowledge selectedShapeKnowledge = null;
	
	private KnowledgeBase loadedKb = null;
	
	private AppKnowledgeBase appKB;
	
	public SketchApp(CyberNode node, GuiAdapter gui, FileAdapter fa){
		super(node);
		this.gui = gui;
		this.fa = fa;
		
		pressTimer = null;
		
		initQueues();	
				
		nodeQuadtree = new ObjectQuadtree(QUDTREE_WIDTH, QUADTREE_HEIGHT, QUADTREE_NODE_CAPACITY);
		edgeQuadtree = new ObjectQuadtree(QUDTREE_WIDTH, QUADTREE_HEIGHT, QUADTREE_NODE_CAPACITY);
		
		initAppKB();
		//System.out.println("BOLCK_SIZ = "+BLOCK_SIZ);
	}
	
	@Override
	public void initialize(){
		postEvent(new SessionEvent(SessionEvent.TYPE.SHOW, this));
	}
	
	@Override
	public void handleEvent(ApplicationEvent e){
		
		// GUI Event from GUI adapter
		if(e instanceof PressEvent){
			switch(((PressEvent) e).action){
			case PRESS:
				press((PressEvent)e);
				break;
			case RELEASE:
				release((PressEvent)e);
				break;
			case DRAG:
				drag((PressEvent)e);	
				break;
			}
		}
		
		else if(e instanceof KeyEvent)
			key(((KeyEvent) e).unicodechar);
		
		else if(e instanceof PopupMenuEvent)
			handlePopupMenuEvent((PopupMenuEvent)e);
		
		else if(e instanceof FilePathEvent)
			handleFilePathEvent((FilePathEvent)e);
		
		else if(e instanceof MapFileEvent){
			if(MapFileEvent.TYPE.SAVE == ((MapFileEvent)e).type)
				saveMap(((MapFileEvent) e).path);
			else
				loadMap(((MapFileEvent) e).path);
		}
		
		else if(e instanceof KnowledgeDisseminationEvent)
			//handleDisseminationEvent((DisseminationEvent)e);
			postKnowledge((ShapeKnowledge)((KnowledgeDisseminationEvent)e).k);
		
		else if(e instanceof KnowledgeAcceptanceEvent)
			addShapeKnowledge((ShapeKnowledge)((KnowledgeAcceptanceEvent)e).k);
		
		else if(e instanceof ErrorMsgEvent)
			gui.popupMsg(((ErrorMsgEvent)e).msg);
		
		else if(e instanceof SessionEvent)
			handleSessionEvent((SessionEvent)e);
		
		else if(e instanceof RestoreEvent)
			handleKnowledge(((RestoreEvent) e).k);
	}
	
	private void handleSessionEvent(SessionEvent e){
		switch(e.type){
		case RESPONSE:
			if(null == session)
				gui.updateSessions(e.sessions);
			break;
			
		case CREATE:
			initGuiAdapter(e.session);
			gui.popupMsg("Session "+e.session.title);
			break;
			
		case SHOW:
			session = null;
			postEvent(new SessionEvent(SessionEvent.TYPE.REQUEST, null, false, SESSION_LIST_UPDATE_DELAY, BrainmapRealWorld.brainmapApp));
			gui.sessionbrowser();	
			break;
			
		case JOIN:
			initGuiAdapter(e.session);
			gui.popupMsg("Session "+e.session.title);
			break;
			
		case LOAD:
			handleSessionLoadEvent(e.session, e.isLoaded);
			break;
		}
	}
	
	private void handleSessionLoadEvent(Session s, boolean isLoaded){
		if(!isLoaded){
			gui.failedToLoadSession("Session \""+s.title+"\" exists");
			return;
		}
		
		session = s;
		initGuiAdapter(session);
		
		for(ApplicationKnowledge k: loadedKb)
			postKnowledge(k);
		
		loadedKb = null;
		gui.successToLoadSession();
	}
	
	private void initGuiAdapter(Session s){
		session = s;
		gui.removeAllShapeKnowledge();
		
		nodeQuadtree = new ObjectQuadtree(QUDTREE_WIDTH, QUADTREE_HEIGHT, QUADTREE_NODE_CAPACITY);
		edgeQuadtree = new ObjectQuadtree(QUDTREE_WIDTH, QUADTREE_HEIGHT, QUADTREE_NODE_CAPACITY);
		
		for(ApplicationKnowledge k: node.kb()){
			if(k instanceof ShapeKnowledge){
				if(((ShapeKnowledge) k).sessionId.equals(s.id))
					addShapeKnowledge((ShapeKnowledge) k);
			}
		}
		
		gui.showDrawingPane("Session: "+s.title);
	}
	
	@Override
	public void handleKnowledge(ApplicationKnowledge k){		
		if(k instanceof ShapeKnowledge || k instanceof ContentKnowledge)
			appKB.updateKnowledge(k);
	}
	
	@Override
	public void postKnowledge(ApplicationKnowledge k){
		//System.out.printf("Post-SketchApp %s\n", k.toString());
		
		if(k instanceof ShapeKnowledge)
			postEvent(new CDRequestEvent(k,((ShapeKnowledge)k).id, BLOCK_SIZ));
			
		else if(k instanceof ContentKnowledge)
			postEvent(new CDRequestEvent(k,((ContentKnowledge)k).id, BLOCK_SIZ));
			
		else
			super.postKnowledge(k);
	}
	
	@Override
	public String name(){
		return super.name()+"::SketchApp";
	}
	
	public void addShapeKnowledge(ShapeKnowledge k){
		//System.out.println("ShapeKnowledge "+k+" is added.");
		
		if(!gui.addShapeKnowledge(k))
			return;
		
		if(k.s instanceof Line){
			//System.out.println("Add Edge!");
			for(Point2df p : k.s.points)
				edgeQuadtree.insert(k.s, p);
			
		}else{	//ClosedShape or FileContentShape
			//System.out.println("Add Node!");
			for(Point2df p : k.s.points)
				nodeQuadtree.insert(k.s, p);
		}
		
	}
	
	public void removeShapeKnowledge(ShapeKnowledge k){
		//System.out.println("SketchApp::removeShapeKnowledge:Remove Shape Knowledge "+k);
		
		// Remove corresponding shape
		gui.removeShapeKnowledge(k);
		
		if(k.s instanceof Line)
			edgeQuadtree.delete(k.s);
		else //ClosedShape or FileContentShape
			nodeQuadtree.delete(k.s);
		
	}
	
	private void press(PressEvent e){
		Point2df p = gui.projectToMap(e.point);
		
		// Press Left-Button
		if(PressEvent.TYPE.LEFT == e.type){
			if(isTexting)
				postEvent(new KeyEvent(10));	//post ENTER key
			activatePressTimer();
			lastPressPoint = p;	
		}
		
		// Press Right-Button
		else if(PressEvent.TYPE.RIGHT == e.type)
			releasePressTimer();
	}
	
	private void release(PressEvent e){
		releasePressTimer();
		Point2df p = gui.projectToMap(e.point);
		
		if(isMoving){
			isMoving = false;
			gui.finishToMoveScreen();
		}
		
		// Release Left-Button
		else if(PressEvent.TYPE.LEFT == e.type){
			if(isDrawing){
				isDrawing = false;
				gui.removeTempShape();
				generateShape(tmpPoints, tmpPointsPointer);
			}else{
				if(System.currentTimeMillis()-lastClickTime < DOUBLE_CLICK_TIME){
					// It's a double click. Enable key input
					lastClickTime = 0;
					gui.enableKeyInput();
					lastDoubleClickPoint = p;
				}else{
					// It's a click.
					lastClickTime = System.currentTimeMillis();
					if(longPress)
						longClick(p);
					else
						click(p);				
				}
			}
		}
	}
	
	private void drag(PressEvent e){
		Point2df p = e.point;
		if(null == lastPressPoint)
			lastPressPoint = gui.projectToMap(p);
		
		// Check Click tolerance
		double dis = lastPressPoint.length(gui.projectToMap(p));
		if(null != pressTimer &&  dis < CLICK_TOLERANCE)
			return;
		
		releasePressTimer();
		
		// Drag Left-Button : Draw line
		if(PressEvent.TYPE.LEFT == e.type){
			if(!isMoving){
				p = gui.projectToMap(p);
				
				if(!isDrawing){
					isDrawing = true;
					tmpPointsPointer = 0;
					tmpPoints = new Point2df[TMP_SHAPE_SIZE_UNIT];
				}else
					gui.removeTempShape();
				
				addPointToTmpPoints(p);
				Line tmpLine = new Line(Color.RED, false, false, tmpPoints, tmpPointsPointer);
				gui.addTempShape(tmpLine);
			}
		}
		
		// Drag Right-Button : Move screen
		else if(PressEvent.TYPE.RIGHT == e.type){
			if(!isDrawing){
				if(!isMoving){
					isMoving = true;
					movingStart = p;
					gui.startToMoveScreen();
				}
				
				Point2df movingVector = new Point2df(movingStart.x-p.x, p.y-movingStart.y);
				gui.moveScreen(movingVector);
			}
		}
	}
	
	private void key(int key){
		releasePressTimer();
		//System.out.println("KEY!!! "+key);
		
		switch(key){
		case BS:{//bs
			if(tmpText.length() > 1)
				tmpText = tmpText.substring(0, tmpText.length()-1);
			else
				tmpText = tmpText.substring(0, 0);
			
			gui.removeTempShape();
			Point2df[] points = new Point2df[1];
			points[0] = lastDoubleClickPoint;
			gui.addTempShape(new Text(tmpText, getEngineColor(), points));
			break;
		}
			
		case ENTER://enter
			if(isTexting){
				//System.out.println("ENTER-TEXTING");
				if(null!=tmpText && 0!=tmpText.compareTo("")){
					editShapeKnowledgeText(lastDoubleClickPoint, tmpText, getEngineColor());					
					tmpText="";
				}
				isTexting = false;
				gui.removeTempShape();
			}
			
			gui.disableKeyInput();
			break;
			
		//case SPACE:	//space
		default:{
			if(isTexting){
				//System.out.println("KEY"+key+"-TEXTING");
				if(gui.shapeNum()-1 > 0)
					gui.removeTempShape();
			}else
				isTexting = true;
			
			tmpText = tmpText+(char)key;
			Point2df[] points = new Point2df[1];
			points[0] = lastDoubleClickPoint;
			gui.addTempShape(new Text(tmpText, getEngineColor(), points));
			break;
			}
		}
	}
	
	private void click(Point2df p){
		
		// Select a ShapeKnowledge on the map
		Shape s = (Shape)nodeQuadtree.select(p, true);
		
		if(null==s || !s.contains(p))
			s = (Shape)edgeQuadtree.select(p, false);
		
		if(null != s){
			//System.out.println("SELECT::"+s);
			selectedShapeKnowledge = gui.selectShapeKnowledge(s.id);
			gui.popupmenu(p);
		}
	}
	
	private void longClick(Point2df p){}
	
	private void longPress(Point2df p){
		lastLongPressPoint = p;
		
		// Attach a file on the map
		gui.attachFile();
	}
	
	private void handleFilePathEvent(FilePathEvent e){
		File file = new File(e.abspath);
		
		if(!file.isFile()){
			gui.popupMsg("No File, "+e.abspath);
			return;
		}
		
		Point2df[] center = new Point2df[1];
		center[0] = lastLongPressPoint;
		FileURI uri = new FileURI(file.getName(), file.length(), file.lastModified());
		FileContent content = FileContent.fileContentFactory(uri, e.abspath);
		
		if(null != content){
			FileContentShape shape = new FileContentShape(uri, getEngineColor(), center);	//Node
			
			if(postNodeShapeEvent(shape, true, true, false))
				postKnowledge(new ContentKnowledge(content, this));
			
		}else
			gui.popupMsg("File size exeeds "+((double)FileContent.MAX_FILE_SIZE/(double)(1024*1024))+"MB");
		
		gui.successToAttachFile();
	}
	
	private void deleteKnowledge(ShapeKnowledge sk){
		postKnowledge(new DeleteKnowledge(sk, this));
		postEvent(new GraphDeleteEvent(sk.id));
	}
	
	private void handlePopupMenuEvent(PopupMenuEvent e){
		if(PopupMenuEvent.menus[0].equals(e.menu))
			deleteKnowledge(selectedShapeKnowledge);
		
		else if(e.menu.startsWith("Edit Color")){	// Edit
			String[] tokens = e.menu.split(PopupMenu.MENU_SEPARATOR);
			if(tokens.length > 1){
				String color = tokens[1];
				for(Color c: Color.values()){
					if(c.toString().equals(color)){
						
						// Change the color of the ShapeKnowledge
						if(null != selectedShapeKnowledge){
							editShapeKnowledgeColor(selectedShapeKnowledge, c);
						}
					}
				}
			}
		}
	}
	
	private void editShapeKnowledgeColor(ShapeKnowledge k, Color c){
		if(k.s instanceof Line){
			Line l = new Line((Line)k.s, c);
			postEdgeShapeEvent(l, false, true);
		}
		
		else if(k.s instanceof ClosedShape){
			ClosedShape cs = new ClosedShape((ClosedShape)k.s, c);
			postNodeShapeEvent(cs, false, false, true);
		}
		
		else if(k.s instanceof FileContentShape){
			FileContentShape fcs = ((FileContentShape)k.s).changeColor((FileContentShape)k.s, c);
			postNodeShapeEvent(fcs, false, false, true);
		}
	}
	
	private void editShapeKnowledgeText(Point2df center, String text, Color color){
		Shape s = (Shape)nodeQuadtree.select(center, ClosedShape.class);
		
		if(null != s && s instanceof ClosedShape){
			if(s.contains(center))
				postNodeShapeEvent(new ClosedShape((ClosedShape)s, text, center, color), false, false, true);
			
			else
				gui.popupMsg("Draw a Node first");
		}
		else
			gui.popupMsg("Draw a Node first");
	}
	
	
	private synchronized void handleShapeKnowledge(ShapeKnowledge k){
		Shape s = k.s;	
		if(s instanceof ContentShape)	
			shapeWaitingQueue.add(k);	// Pairing with ContentKnowledge	
		else
			//postEvent(new DisseminationEvent(k, DisseminationEvent.SOURCE.GLOBAL, BrainmapRealWorld.brainmapApp));
			postEvent(new KnowledgeAcceptanceEvent(k, BrainmapRealWorld.brainmapApp));
	}
	
	private synchronized void handleContentKnowledge(ContentKnowledge k){			
		k.c.store(fa);
		contentWaitingQueue.add(k);	// Pairing with ShapeKnowledge 
	}	
	
	private void generateShape(Point2df[] rawPoints, int len){
		// Generate Shape
		Shape s = null;
		len = removeDuplicatedPoint(rawPoints, len);
		Point2df[] points = this.getClosedShape(rawPoints, len);
		
		if(null == points){	//Edge
			points = new Point2df[len];
			for(int i=0; i<len; i++)
				points[i] = (Point2df)(rawPoints[i]);
			
			s = new Line(getEngineColor(), false, false, points, len);
		}
		
		else{	//Node
			s = new ClosedShape(Color.NONE, getEngineColor(), points, points.length);
			
			// Check area of ClosedShape
			if(((ClosedShape)s).area() < MIN_CLOSED_SHAPE_AREA){
				gui.popupMsg("Node is Too small");
				return;
			}
				
		}
		
		// Map the Shape with corresponding shapes
		if(s instanceof Line)
			postEdgeShapeEvent((Line)s, true, false);
		
		else if(null != s)
			postNodeShapeEvent(s, true, true, false);
	}
	
	private void postEdgeShapeEvent(Line s, boolean checkDuplication, boolean skOnly){
		String n1 = null;
		Shape s1 = (Shape)nodeQuadtree.select(s.points[0], true);
		if(null != s1 && s1.contains(s.points[0])){
			s.cut(true, s1);
			s.setHead(true);
			n1 = s1.id;			
		}
		
		String n2 = null;
		Shape s2 = (Shape)nodeQuadtree.select(s.points[s.length-1], true);
		if(null != s2 && s2.contains(s.points[s.length-1])){
			if(s2.contains(s)){
				gui.popupMsg("Invalid Edge");
				return;
			}
			
			s.cut(false, s2);
			s.setTail(true);
			n2 = s2.id;
		}
		
		postEvent(new ShapeEvent(s, n1, n2, checkDuplication, skOnly));
	}
	
	private boolean postNodeShapeEvent(Shape s, boolean checkConfliction, boolean checkEdges, boolean editShapeOnly){
		HashSet<String> conflictionChecked = null;
		HashSet<String> connectedEdges = null;
		HashSet<String> connectedNodes = null;
		
		if(checkConfliction)
			conflictionChecked = new HashSet<String>();
		
		if(checkEdges){
			connectedEdges = new HashSet<String>();
			connectedNodes = new HashSet<String>();
		}
		
		for(int i=0; i<s.length; i++){
			// Check confliction
			if(checkConfliction){
				Shape s1 = (Shape)nodeQuadtree.select(s.points[i], true);
				if(null != s1 && !conflictionChecked.contains(s1.id)){
					conflictionChecked.add(s1.id);
					
					if(s.conflictWith(s1)){
						gui.popupMsg("Conflict with other nodes!");
						return false;
					}
				}
			}
			
			// Check connected edges
			if(checkEdges){
				Line s2 = (Line)edgeQuadtree.select(s.points[i], true);
				
				// Connected to head
				if(null != s2 && (s.contains(s2.points[0]))){	
					if(!connectedEdges.contains(s2.id)){
						connectedEdges.add(s2.id);
						s2.cut(true, s);
						s2.setHead(true);
						
						Shape n2 = (Shape)nodeQuadtree.select(s2.points[s2.length-1], true);
						
						if(null == n2 || n2 == s){
							postEvent(new ShapeEvent(s2, s.id, null, true, false));
							continue;
						}
						
						if(connectedNodes.contains(n2.id)){
							deleteKnowledge(gui.selectShapeKnowledge(s2.id));
							gui.popupMsg("Remove Duplication");
							continue;
						}
						
						String n2_id = n2.id;
						connectedNodes.add(n2_id);
						postEvent(new ShapeEvent(s2, s.id, n2_id, true, false));
					}
				}
				
				// Connected to tail
				else if(null != s2 && s.contains(s2.points[s2.length-1])){
					if(!connectedEdges.contains(s2.id)){
						connectedEdges.add(s2.id);
						s2.cut(false, s);
						s2.setTail(true);
						
						Shape n2 = (Shape)nodeQuadtree.select(s2.points[0], true);
						
						if(null == n2 || n2 == s){
							postEvent(new ShapeEvent(s2, s.id, null, true, false));
							continue;
						}
						
						if(connectedNodes.contains(n2.id)){
							deleteKnowledge(gui.selectShapeKnowledge(s2.id));
							gui.popupMsg("Remove Duplication");
							continue;
						}
						
						String n2_id = n2.id;
						connectedNodes.add(n2_id);
						postEvent(new ShapeEvent(s2, s.id, n2_id, true, false));
					}
				}
			}
		}
		postEvent(new ShapeEvent(s, null, null, true, editShapeOnly));
		return true;
	}
	
	private void addPointToTmpPoints(Point2df p){
		// Increase array size
		if(tmpPoints.length <= tmpPointsPointer){
			Point2df[] newArray = new Point2df[tmpPoints.length + TMP_SHAPE_SIZE_UNIT];
			
			for(int index=0; index<tmpPoints.length; index++)
				newArray[index] = tmpPoints[index];
			tmpPoints = newArray;
		}
		
		tmpPoints[tmpPointsPointer++] = p;
	}
	
	private void loadMap(String path){
		try{			
			ObjectInputStream ois = new ObjectInputStream(new FileInputStream(path));
			loadedKb = (KnowledgeBase)ois.readObject();
			
			ois.close();
			
			for(ApplicationKnowledge k: loadedKb){
				if(k instanceof SessionKnowledge)
					postEvent(new SessionEvent(SessionEvent.TYPE.LOAD,((SessionKnowledge) k).session, true, 0.0, BrainmapRealWorld.brainmapApp));
			}
			
		}catch(Exception e){
			gui.failedToLoadSession("Failed to Load Session");
		}
	}
	
	private void saveMap(String path){
		try{
			FileOutputStream fos = new FileOutputStream(path);
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			ObjectOutputStream oos = new ObjectOutputStream(bos);
			
			KnowledgeBase kb = new KnowledgeBase();
			
			// Filter current session's knowledge only
			for(ApplicationKnowledge k : node.kb()){
				if(k instanceof ShapeKnowledge){
					if(((ShapeKnowledge)k).sessionId.equals(session.id))
						kb.add(k);
				}
				
				else if(k instanceof SessionKnowledge){
					if(((SessionKnowledge)k).id.equals(session.id)){
						
						Session s = ((SessionKnowledge)k).session;
						s.removeAllEngines();
						
						kb.add(k);
					}
				}
				
				else if(k instanceof DeleteKnowledge || k instanceof GraphKnowledge 
						|| k instanceof BlockKnowledge || k instanceof ContentKnowledge){
					kb.add(k);
				}
			}
			
			oos.writeObject(kb);
			oos.flush();
			
			byte[] bytes = bos.toByteArray();
			
			fos.write(bytes);
			fos.flush();
			
			oos.close();
			bos.close();
			fos.close();
			
			gui.hideDrawingPane();
			postEvent(new SessionEvent(SessionEvent.TYPE.EXIT, BrainmapRealWorld.brainmapApp));
			postEvent(new SessionEvent(SessionEvent.TYPE.SHOW, this));
			gui.successToSaveSession();
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	private synchronized void activatePressTimer(){
		releasePressTimer();
		pressTimer = new Timer();
		pressTimer.schedule(new TimerTask(){
			public void run(){
				longPress=true;
				longPress(lastPressPoint);
			}
		}, LONG_CLICK_TIME);
	}
	
	private synchronized void releasePressTimer(){
		if(null!=pressTimer){
			pressTimer.cancel();
			pressTimer = null;
		}
		longPress=false;
	}
	
	// generate random color except white and red
	private Color getEngineColor(){
		int engineid = Math.abs((int)BrainmapRealEngine.myEngineId());
		//System.out.println("Engine id = " + engineid);
		//System.out.println("Length = " + Color.values().length);
		
		return Color.values()[engineid%(Color.values().length-3)];
	}
	
	@Override
	public void onMet(Waitable w1, Waitable w2){
		ShapeKnowledge s = null;
		ContentKnowledge c = null;
		
		if(w1 instanceof ShapeKnowledge && w2 instanceof ContentKnowledge){
			s = (ShapeKnowledge)w1;
			c = (ContentKnowledge)w2;
		}
		
		else if(w2 instanceof ShapeKnowledge && w1 instanceof ContentKnowledge){
			s = (ShapeKnowledge)w2;
			c = (ContentKnowledge)w1;
		}
		
		else{
			System.err.println("SketchApp::onMet:FATAL : both of Waitables are same type.. "+w1.getClass().getName()+", "+w2.getClass().getName());
			return;
		}
		
		//postEvent(new DisseminationEvent(s, DisseminationEvent.SOURCE.GLOBAL, BrainmapRealWorld.brainmapApp));
		postEvent(new KnowledgeAcceptanceEvent(s, BrainmapRealWorld.brainmapApp));
	}
	
	@Override
	public void onHistory(Waitable w){
		if(w instanceof ShapeKnowledge)
			//postEvent(new DisseminationEvent((ShapeKnowledge)w, DisseminationEvent.SOURCE.GLOBAL, BrainmapRealWorld.brainmapApp));
			postEvent(new KnowledgeAcceptanceEvent((ShapeKnowledge)w, BrainmapRealWorld.brainmapApp));
		
		else if(w instanceof ContentKnowledge){}
	}
	
	// Cut down points after first intersection point
	private Point2df[] getIntersectedClosedShape(Point2df[] shape, int length){
		for(int i=1; i<length; i++){
			Point2df p1=shape[i-1];
			Point2df p2=shape[i];
			float minx = p1.x<p2.x? p1.x:p2.x;
			float maxx = p1.x<p2.x? p2.x:p1.x;
			float miny = p1.y<p2.y? p1.y:p2.y;
			float maxy = p1.y<p2.y? p2.y:p1.y;
			
			for(int j=i+2; j<length; j++){
				Point2df p3=shape[j-1];
				Point2df p4=shape[j];
				float minx2 = p3.x<p4.x? p3.x:p4.x;
				float maxx2 = p3.x<p4.x? p4.x:p3.x;
				if((minx<=p3.x && p3.x <= maxx) || (minx<=p4.x && p4.x<=maxx) || (minx2<minx && maxx<maxx2)){
					float miny2 = p3.y<p4.y? p3.y:p4.y;
					float maxy2 = p3.y<p4.y? p4.y:p3.y;
					if((miny<=p3.y && p3.y <= maxy) || (miny<=p4.y && p4.y<=maxy) || (miny2<miny && maxy<maxy2)){
						float dx1 = p2.x-p1.x;
						float dx2 = p4.x-p3.x;
						float dy1 = p2.y-p1.y;
						float dy2 = p4.y-p3.y;
						float m1=(dx1==0)? Float.MAX_VALUE:dy1/dx1;
						float m2=(dx2==0)? Float.MAX_VALUE:dy2/dx2;
						float x1=p1.x;
						float x2=p3.x;
						float y1=p1.y;
						float y2=p3.y;
						float x = Float.NaN;
						float y = Float.NaN;
						if(0 == m1){
							if(0 == m2){
								if(y1 == y2){
									x = ((x1+dx1/2.0f)+(x2+dx2/2.0f)) / 2.0f;	//ave of ave
									y = y1;
								}								
							}else if(Float.MAX_VALUE == m2){
								x = x2;
								y = y1;
							}else{
								y = y1;
								x = (y-y2)/m2+x2;
							}
						}else if(Float.MAX_VALUE == m1){ 
							if(0 == m2){
								x = x1;
								y = y2;
							}else if(Float.MAX_VALUE == m2){
								if(x1==x2){
									x = x1;
									y = ((y1+dy1/2.0f)+(y2+dy2/2.0f)) / 2.0f;
								}
							}else{
								x = x1;
								y = m2*(x-x2)+y2;
							}
						}else{
							if(0 == m2){
								y = y2;
								x = (y-y1)/m1+x1;
							}else if(Float.MAX_VALUE == m2){
								x = x2;
								y = m1*(x-x1)+y1;
							}else{
								if(m1 == m2){	//paralell		
									if(y2 == m1*(x2-x1)+y1){
										x = ((x1+dx1/2.0f)+(x2+dx2/2.0f)) / 2.0f;	//ave of ave
										y = ((y1+dy1/2.0f)+(y2+dy2/2.0f)) / 2.0f;
									}
								}else{
									x = (m1*x1-m2*x2-y1+y2)/(m1-m2);
									y = m1*(x-x1)+y1;
								}
							}
						}
						if(Float.NaN == x || Float.NaN == y)
							continue;
						
						// intersection range
						minx = minx<minx2? minx:minx2;
						maxx = maxx<maxx2? maxx2:maxx;
						miny = miny<miny2? miny:miny2;
						maxy = maxy<maxy2? maxy2:maxy;
						
						// Cut
						if((minx<=x && x<=maxx) && (miny<=y && y<=maxy)){
							Point2df[] result = new Point2df[j-i+2];
							result[0] =new Point2df(x, y); 
							
							for(int index=i; index<j; index++)
								result[index-i+1] = shape[index];
							result[j-i+1] = new Point2df(x, y);
							
							return result;
						}
					}
				}
			}
		}
		return null;
	}
	
	private Point2df[] getClosedShape(Point2df[] shape, int length){
		// is there an intersection point?
		if(length < 5)
			return null;
		
		Point2df[] intersectedShape = this.getIntersectedClosedShape(shape, length);
		if(null != intersectedShape)
			return intersectedShape;
		
		// if there is no intersection point, measure the length 
		// if the length is  not less than total length/10, return null.
		double len = shape[0].length(shape[length-1]);
		double totLen = 0.0;
		
		for(int index=1; index<length; index++)
			totLen += shape[index-1].length(shape[index]);
		
		if(len < totLen/10.0){
			Point2df[] extendedShape = new Point2df[length+1];
			
			for(int index=0; index<length; index++)
				extendedShape[index] = shape[index];
			
			extendedShape[length] = shape[0];
			return extendedShape;
		}
		return null;
	}
	
	// Returns : length of new points
	private int removeDuplicatedPoint(Point2df[] points, int len){
		for(int i=1; i<len; i++){
			if(points[i-1] == points[i])
				points[i] = null;
			else if(null == points[i-1]){
				Point2df pre = null;
				int j = 2;
				while(i>j-1 && null==pre){
					pre = points[i-j];
					j++;
				}
				if(null == pre)
					System.err.println("Failed to remove duplicated points : previous point is null");
				else if(pre == points[i])
					points[i] = null;
			}
		}
		
		int count=1;
		
		for(int i=1; i<len; i++){
			if(null != points[i])
				points[count++] = points[i];
		}
		
		return count;
	}
	
	private void initQueues(){
		shapeWaitingQueue = new WaitingQueue("Shape", true);
		contentWaitingQueue = new WaitingQueue("Content", false);
		
		shapeWaitingQueue.setWaitingQueue(contentWaitingQueue);
		contentWaitingQueue.setWaitingQueue(shapeWaitingQueue);
		
		shapeWaitingQueue.setWaitingQueueListener(this);
		contentWaitingQueue.setWaitingQueueListener(this);
	}
	
	private void initAppKB(){
		appKB = new AppKnowledgeBase(){
			
			public void handleKnowledge(ApplicationKnowledge k){
				if(k instanceof ShapeKnowledge){
					System.out.println("Received-SketchApp "+k);
					handleShapeKnowledge((ShapeKnowledge)k);
					
				}else if(k instanceof ContentKnowledge){
					//System.out.println("Received-SketchApp "+k);
					handleContentKnowledge((ContentKnowledge)k);
				}
			}
		};
	}
}

class ContentKnowledge extends EditableKnowledge implements Waitable{
	private static final long serialVersionUID = 1L;
	public Content c;
	public String id;
	
	public ContentKnowledge(Content c, CyberApplication app){
		super(c.uri.id(), app);
		this.c = c;
		this.id = c.uri.id();
	}
	
	@Override
	public String toString(){
		return "ContentKnowledge-"+c.id()+" ("+c+")";
	}
	
	public String id() {
		return id;
	}
}
