/*

   Copyright 2010, 2013 SRI International

   Licensed under the Apache License, Version 2.0 (the "License"); you may
   not use this file except in compliance with the License. You may obtain
   a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

   The development was supported by National Science Foundation Grant 0932397:
   A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package com.sri.brainmap;

import java.io.File;
import java.io.BufferedReader;
import java.io.InputStreamReader;

import com.sri.brainmap.R;

import android.app.Activity;
import android.content.Context;
import android.content.Intent;
import android.os.Bundle;
import android.view.LayoutInflater;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.ViewGroup;
import android.widget.AdapterView;
import android.widget.AdapterView.OnItemClickListener;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.EditText;
import android.widget.ImageView;
import android.widget.LinearLayout;
import android.widget.ListView;
import android.widget.TextView;
import android.widget.Toast;

public class FileBrowserActivity extends Activity implements OnItemClickListener, OnClickListener{
	
	// File Browser Results
	public static enum RESULT {CANCEL, FILE_SELECTED};
	
	private final static String ROOT="/";
	private final static String PARENT="..";
	
	// Views
	ListView lv;
	EditText et;
	Button button;
	
	// Initial Location
	String currentPath="/sdcard";
	
	@Override
	public void onCreate(Bundle savedInstanceState){
		super.onCreate(savedInstanceState);
		initialize();
		refreshList(currentPath);
	}
	
	public void initialize(){
		setContentView(R.layout.filebrowser);
		
		lv = (ListView)findViewById(R.id.listView1);
		et = (EditText)findViewById(R.id.editText1);
		button = (Button)findViewById(R.id.button1);
		
		lv.setOnItemClickListener(this);
		button.setOnClickListener(this);
	}
	
	public void popupToast(String msg){
		Toast.makeText(this, msg.subSequence(0, msg.length()), Toast.LENGTH_SHORT).show();
	}
	
	@Override
	public void onClick(View v) {
		String filename = currentPath+File.separator+et.getText().toString();
		returnPath(filename);
	}
	
	@Override
	public void onItemClick(AdapterView<?> parent, View view, int position, long id){
		String filename = ((TextView)view.findViewById(R.id.tv)).getText().toString();
		String tmpPath;
		
		if(!currentPath.equals(ROOT))
			tmpPath = currentPath+File.separator+filename;
		else
			tmpPath = currentPath+filename;
		
		// Remove ".." from the path.
		if(tmpPath.endsWith(PARENT)){
			tmpPath = tmpPath.substring(0, tmpPath.length()-PARENT.length()-1);
			String substr = tmpPath;
			
			int index = substr.indexOf('/');
			while(-1 != index){
				substr = substr.substring(index+1);
				index = substr.indexOf('/');
			}
			
			tmpPath = tmpPath.substring(0, tmpPath.indexOf(substr)-1);
			if(tmpPath.equals(""))
				tmpPath = ROOT;
		}
		
		// Check permission of file/directory
		if(isAccessible(tmpPath)){
			currentPath = tmpPath;
			File file = new File(currentPath);
			
			if(file.isDirectory())
				refreshList(currentPath);
			else if(file.isFile())
				returnPath(currentPath);
			else
				System.err.println("FileBrowserActivity::onItemClick:Invalid Path!:"+currentPath);
		}
		
		else
			popupToast("Permission Denied.");
	}
	
	// Return : Returns the file permission. 
	//          If there is an exception or the path is invalid, returns -1.
	private int getFilePermission(String path){
		try{
			String cmd = "ls -dl "+path;
			Process p = Runtime.getRuntime().exec(cmd);
			
			BufferedReader br = new BufferedReader(new InputStreamReader(p.getInputStream()));
			String ret="";
			String line;
			
			while((line= br.readLine())!=null)
				ret+=line;
			
			p.waitFor();
			ret.trim();
			int permission = 0;
			
			for(int index=9; index>0; index--){
				if('-' != (char)ret.charAt(index))
					permission += (int)Math.pow(2.0, 9.0-(double)index);
			}
				
			if(permission < 0 || permission > 0777)
				return -1;
			
			return permission;
		}
		
		catch(Exception e){
			e.printStackTrace();
			return -1;
		}
	}
	
	private boolean isAccessible(String path){
		int permission = getFilePermission(path);
		if(-1 == permission)
			return false;
		
		if(permission % 010 < 04)	//check 'r' permission of 'Others'
			return false;
		return true;
	}
	
	@Override
	public void onBackPressed(){
		Bundle extra = new Bundle();
		Intent intent = new Intent();
		extra.putSerializable("Result", RESULT.CANCEL);
		intent.putExtras(extra);
		
		setResult(Activity.RESULT_OK, intent);
		finish();
	}
	
	public void returnPath(String path){
		Bundle extra = new Bundle();
		Intent intent = new Intent();
		extra.putSerializable("Result", RESULT.FILE_SELECTED);
		extra.putString("Path", path);
		intent.putExtras(extra);
		
		setResult(Activity.RESULT_OK, intent);
		finish();
	}
	
	public void refreshList(String dirpath){
		File dir = new File(dirpath);
		if(!dir.isDirectory()){
			//System.out.println("DIRPATH is not the directory!!!");
			return;
		}
		
		File[] files = dir.listFiles();
		File[] filelist;
		if(dirpath.equals(ROOT))
			filelist = files;
		else{
			filelist = new File[files.length+1];
			filelist[0] = new File(dirpath+File.separator+PARENT);
			for(int index=0; index<files.length; index++)
				filelist[index+1] = files[index];
		}
		
		lv.setAdapter(new FileArrayAdapter(this, filelist));
	}
}

class FileArrayAdapter extends ArrayAdapter<File>{
	private final static int RESOURCE_ID = R.layout.filerow;
	private Context context;
	private File[] files;
	
	public FileArrayAdapter(Context context, File[] files) {
		super(context, RESOURCE_ID, files);
		this.context = context;
		this.files = files;
	}
	
	public View getView(int position, View convertView, ViewGroup parent){
		LayoutInflater inflater = (LayoutInflater)context.getSystemService(Context.LAYOUT_INFLATER_SERVICE);
		View rowView = inflater.inflate(RESOURCE_ID, parent, false);
		
		TextView tv = (TextView)rowView.findViewById(R.id.tv);
		ImageView iv = (ImageView)rowView.findViewById(R.id.iv);
		
		File file = files[position];
		tv.setText(file.getName());
		
		if(file.isDirectory())
			iv.setImageResource(R.drawable.dir);
		
		else if(file.isFile())
			iv.setImageResource(R.drawable.file);
		
		else
			iv.setImageResource(R.drawable.question);
		
		return rowView;
	}
}
