/*

   Copyright 2012-2013 SRI International

   Licensed under the Apache License, Version 2.0 (the "License"); you may
   not use this file except in compliance with the License. You may obtain
   a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

   The development was supported by National Science Foundation Grant 0932397:
   A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.app.brainmap.jre;

import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.Spring;
import javax.swing.SpringLayout;

import cf.app.brainmap.ALIGN;
import cf.app.brainmap.BrainmapRealEngine;
import cf.app.brainmap.BrainmapRealWorld;
import cf.app.brainmap.Color;
import cf.app.brainmap.GuiAdapter;
import cf.app.brainmap.ObjectQuadtree;
import cf.app.brainmap.Point2df;
import cf.app.brainmap.PopupMenu;
import cf.app.brainmap.Session;
import cf.app.brainmap.Shape;
import cf.app.brainmap.ShapeKnowledge;
import cf.app.brainmap.SketchApp;
import cf.app.brainmap.events.*;
import cf.app.brainmap.Color;

public class JREGuiAdapter extends JPanel implements GuiAdapter, MouseListener, KeyListener, MouseMotionListener, MouseWheelListener, WindowListener{
	private static final int INITIAL_FRAME_WIDTH = 360;
	private static final int INITIAL_FRAME_HEIGHT = 600;
	
	// Painting
	private JFrame frame;
	private FontMetrics fm;
	private Graphics g = null;
	private static final int BS=8;
	private boolean isKeyBoradEnabled;
	private boolean isRightButton=false;
	private HashMap<String,Image> cachedImages;
	
	// Menu
	private PopupMenu popupMenu = null;
	private Color selectedShapeColor = null;
	private ShapeKnowledge selectedShapeKnowledge = null;
		
	// Projection
	private double resolution = 1.0;
	private Point2df startlt = null;
	private Point2df maplt = null;
	
	// Shape and ShapeKnowledge
	private Shape tempShape = null;
	private Hashtable<String, ShapeKnowledge> shapes;
	
	private SessionBrowser sessionBrowser = null;
	
	static {
		System.setProperty("java.awt.headless", "false");
		System.out.println("Headless="+java.awt.GraphicsEnvironment.isHeadless());
	}
	
	public JREGuiAdapter(){
		super();
		cachedImages = new HashMap<String,Image>();
		shapes = new Hashtable<String, ShapeKnowledge>();
		
		initFrame();
		
		fm = getFontMetrics(getFont());
		isKeyBoradEnabled = false;
		
		int width = frame.getSize().width;
		int height = frame.getSize().height;
		maplt = new Point2df(-width/2, height/2);
	}
	
	private void initFrame(){
		frame = new JFrame("BrainMap");
		frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		
		frame.setSize(new Dimension(INITIAL_FRAME_WIDTH, INITIAL_FRAME_HEIGHT));
		int x = (Toolkit.getDefaultToolkit().getScreenSize().width - INITIAL_FRAME_WIDTH) / 2;
		int y = (Toolkit.getDefaultToolkit().getScreenSize().height - INITIAL_FRAME_HEIGHT) / 2;
		frame.setLocation(x, y);
		
		frame.add(this);
		
		addMouseListener(this);
		addMouseWheelListener(this);
		addMouseMotionListener(this);
		frame.addKeyListener(this);
		frame.addWindowListener(this);
	}
	
	@Override
	public void attachFile(){
		filebrowser("Attach");
	}
	
	@Override
	public void failedToLoadSession(String errmsg){
		popupMsg(errmsg);
		
		SketchApp app = BrainmapRealWorld.sketchApp;
		app.postEvent(new SessionEvent(SessionEvent.TYPE.SHOW, app));
	}
	
	@Override
	public void successToLoadSession(){}
	
	@Override
	public void successToSaveSession(){
		popupMsg("Session Saved");
	}
	
	@Override
	public void failedToSaveSession(){
		popupMsg("Failed to Save Session");
	}
	
	@Override
	public void successToAttachFile(){}	
	
	@Override
	public void showDrawingPane(String title){
		frame.setTitle(title);
		
		if(null != sessionBrowser){
			frame.setLocation(sessionBrowser.getLocation());
			sessionBrowser.setVisible(false);
		}
		
		frame.setVisible(true);
	}
	
	@Override
	public void hideDrawingPane(){
		frame.setVisible(false);
	}
		
	@Override
	synchronized public void paintComponent(Graphics g){
		super.paintComponent(g);
		this.g=g;
		for(String key : shapes.keySet()){
			Shape s = shapes.get(key).s;
			if(isVisibleOnScreen(s))
				s.paint(this);
		}
		
		if(null != tempShape){
			if(isVisibleOnScreen(tempShape)){
				tempShape.paint(this);
			}
		}
		
		if(null!=selectedShapeKnowledge)
			selectedShapeKnowledge.s.paint(this);
		
		if(null != popupMenu)
			popupMenu.paint(this);
		
		this.g=null;
	}
	
	@Override
	public void drawLine(Point2df[] points, int length, Color color){
		Graphics g = getCurrentGraphics();
		if(null == g)
			throw new NullPointerException("Current graphics is unavailable.");
		
		g.setColor(shapeColor2JREColor(color));
		Point2df[] projectedPoints = projectToScreen(points, length);
		for(int pointIndex=1; pointIndex<length; pointIndex++){
			Point2df p1 = projectedPoints[pointIndex-1];
			Point2df p2 = projectedPoints[pointIndex];
			g.drawLine((int)p1.x, (int)p1.y, (int)p2.x, (int)p2.y);
		}
	}
	
	@Override
	// Swing draw string from left bottom
	public void drawString(String str, ALIGN align, Point2df p, Color color) {
		Graphics g = getCurrentGraphics();
		if(null == g)
			throw new NullPointerException("Current graphics is unavailable.");
		
		p = projectToScreen(p);
		int width = fm.stringWidth(str);
		int height = fm.getHeight();
		g.setColor(shapeColor2JREColor(color));
		
		switch(align){
			case CENTER:
				Point2df lb = new Point2df(p.x-(float)width/2.0f, p.y-(float)height/2.0f);
				g.drawString(str, (int)lb.x, (int)lb.y);
				break;
				
			case CENTER_BOTTOM:
				g.drawString(str, (int)(p.x-(float)width/2.0f), (int)p.y);
				break;
				
			case LEFT_TOP:
				g.drawString(str, (int)p.x, (int)p.y + height);
				break;
				
			case RIGHT_TOP:
				g.drawString(str, (int)p.x-width, (int)p.y + height);
				break;
		}
	}

	@Override
	public void drawFilledRect(Point2df lt, float width, float height, Color color) {
		Graphics g = getCurrentGraphics();
		if(null == g)
			throw new NullPointerException("Current graphics is unavailable.");
		
		lt = projectToScreen(lt);
		g.setColor(shapeColor2JREColor(color));
		g.fillRect((int)lt.x, (int)lt.y, (int)width, (int)height);
	}

	@Override
	public void drawFilledCircle(Point2df center, float radius, Color color) {
		Graphics g = getCurrentGraphics();
		if(null == g)
			throw new NullPointerException("Current graphics is unavailable.");
		
		center = projectToScreen(center);
		g.setColor(shapeColor2JREColor(color));
		g.fillOval((int)(center.x-radius), (int)(center.y-radius), (int)radius*2, (int)radius*2);
	}
	
	@Override
	public void drawFilledPolygon(Point2df[] points, int length, Color lineColor, Color fillColor){
		Graphics g = getCurrentGraphics();
		if(null == g)
			throw new NullPointerException("Current graphics is unavailable.");
		
		g.setColor(shapeColor2JREColor(lineColor));
		Point2df[] projectedPoints = projectToScreen(points, length);
		
		for(int pointIndex=1; pointIndex<length; pointIndex++){
			Point2df p1 = projectedPoints[pointIndex-1];
			Point2df p2 = projectedPoints[pointIndex];
			g.drawLine((int)p1.x, (int)p1.y, (int)p2.x, (int)p2.y);
		}
		
		g.setColor(shapeColor2JREColor(fillColor));
	}
	
	@Override
	public void drawImage(String path, Point2df center, float width, float height){
		Graphics g = getCurrentGraphics();
		if(null == g)
			throw new NullPointerException("Current graphics is unavailable.");
		
		Image img = cachedImages.get(path);
		
		if(null == img)
			cachingImage(path);
		else{
			center = projectToScreen(center);
			Point2df leftTop = new Point2df(center.x-width/2, center.y-height/2);
			g.drawImage(img, (int)leftTop.x, (int)leftTop.y, (int)width, (int)height, this);
		}		
	}
	
	@Override
	public Point2df projectToScreen(Point2df p){
		return new Point2df(p.x - maplt.x, maplt.y - p.y);
	}	
	
	public Point2df[] projectToScreen(Point2df[] points, int len){
		Point2df[] projected = new Point2df[len];
		for(int i=0; i<len; i++)
			projected[i] = projectToScreen(points[i]);
		return projected;
	}
	
	@Override
	public Point2df projectToMap(Point2df p){
		return new Point2df(maplt.x + p.x, maplt.y - p.y);
	}
	
	private Point2df[] projectToMap(Point2df[] points, int len){
		Point2df[] projected = new Point2df[len];
		for(int i=0; i<len; i++)
			projected[i] = projectToMap(points[i]);
		return projected;
	}
	
	//For GuiAdapter	
	@Override
	public void enableKeyInput() {
		isKeyBoradEnabled = true;
	}

	@Override
	public void disableKeyInput() {
		isKeyBoradEnabled = false;
	}
	
	@Override
	synchronized public boolean addShapeKnowledge(ShapeKnowledge s) {
		//shapes.add(s);
		if(null != shapes.get(s.id))
			return false;
		
		shapes.put(s.id, s);
		this.repaint();
		/*for(Point2df p : s.s.points){
			quadtree.insert(s, p);
		}*/
		return true;
	}
	
	synchronized public void removeAllShapeKnowledge(){
		shapes = new Hashtable<String, ShapeKnowledge>();
	}

	@Override
	synchronized public void removeShapeKnowledge(ShapeKnowledge s) {
		shapes.remove(s.id);
		this.repaint();
	}
	
	@Override
	public ShapeKnowledge getShapeKnowledge(String id){
		return shapes.get(id);
	}
	
	@Override
	public ShapeKnowledge selectShapeKnowledge(String id){
		ShapeKnowledge s = shapes.get(id);
		if(null == s)
			return null;
		
		selectedShapeKnowledge = s.clone();
		selectedShapeColor = s.s.color;
		selectedShapeKnowledge.s.color = Color.RED;
		repaint();
		return s;
	}
	
	@Override
	synchronized public void addTempShape(Shape s){
		tempShape = s;
		this.repaint();
	}
	
	@Override
	synchronized public void removeTempShape(){
		if(null != tempShape){
			tempShape = null;
		}
		this.repaint();
	}
	
	@Override
	public int shapeNum() {
		return shapes.size();
	}
	
	@Override
	public void moveScreen(Point2df vector){
		if(null == startlt)
			return;
		
		maplt = new Point2df(startlt.x+vector.x, startlt.y+vector.y);
		repaint();
	}
	
	@Override
	public void startToMoveScreen(){
		startlt = new Point2df(maplt.x, maplt.y);
	}
	
	@Override
	public void finishToMoveScreen(){
		startlt = null;
	}
	
	public void filebrowser(String selectButtonName) {
		JFileChooser c = new JFileChooser();
		int rval = c.showDialog(frame, selectButtonName);
		SketchApp app = BrainmapRealWorld.sketchApp;
		switch(rval){
		case JFileChooser.APPROVE_OPTION:
			//System.out.println("Selected file path : "+c.getSelectedFile().getAbsolutePath());
			app.postEvent(new FilePathEvent(c.getSelectedFile().getAbsolutePath()));
			break;
			
		case JFileChooser.CANCEL_OPTION:
			break;
			
		default:
			break;
		}
	}
	
	@Override
	public void sessionbrowser(){
		frame.setVisible(false);
		if(null == sessionBrowser)
			sessionBrowser = new SessionBrowser(this, frame.getLocation());
		else
			sessionBrowser.setVisible(true, frame.getLocation());
	}
	
	@Override
	public void updateSessions(ArrayList<Session> sessions){
		if(null != sessionBrowser)
			sessionBrowser.updateSessions(sessions);
	}
	
	@Override
	public void popupMsg(String msg){
		//JOptionPane.showMessageDialog(this, msg);
		JOptionPane.showMessageDialog(this, msg, "Warning", JOptionPane.PLAIN_MESSAGE, null);
		//System.out.println("POPUP!!! "+msg);
	}
	
	public void popupLoadDialog(){
		JFileChooser c = new JFileChooser();
		int rval = c.showDialog(frame, "Load");
		SketchApp app = BrainmapRealWorld.sketchApp;
		
		switch(rval){
			case JFileChooser.APPROVE_OPTION:
				//System.out.println("Selected file path : "+c.getSelectedFile().getAbsolutePath());
				app.postEvent(new MapFileEvent(MapFileEvent.TYPE.LOAD, c.getSelectedFile().getAbsolutePath()));
				break;
		}
	}
	
	@Override
	public void popupmenu(Point2df p){
		int maxWidth = 0;
		for(int i=0; i<PopupMenu.MENU_NUM; i++){
			if(maxWidth < fm.stringWidth(PopupMenuEvent.menus[i]))
				maxWidth = fm.stringWidth(PopupMenuEvent.menus[i]);
		}
		
		int width = maxWidth + PopupMenu.MENU_NUM * PopupMenu.MARGIN;
		int height = (fm.getHeight()+PopupMenu.MARGIN*2)*PopupMenu.MENU_NUM;
		int frameWidth = frame.getSize().width;
		int frameHeight = frame.getSize().height;
		
		popupMenu = new PopupMenu(p, projectToMap(new Point2df(0.0f, 0.0f)), width, height, frameWidth, frameHeight);
		repaint();
	}
	
	private int showMapFileDialog(MapFileEvent.TYPE type, String msg, String buttonName){
		int ret = JOptionPane.showConfirmDialog(this, msg);
		switch(ret){			
			case JOptionPane.YES_OPTION:
				JFileChooser c = new JFileChooser();
				int rval = c.showDialog(frame, buttonName);
				SketchApp app = BrainmapRealWorld.sketchApp;
				
				switch(rval){
					case JFileChooser.APPROVE_OPTION:
						//System.out.println("Selected file path : "+c.getSelectedFile().getAbsolutePath());
						app.postEvent(new MapFileEvent(type, c.getSelectedFile().getAbsolutePath()));
						break;
					default:
						JOptionPane.showMessageDialog(c, "Failed to attach a file.");
						break;
				}
			break;
			
		}
		return ret;
	}
	
	private void closePopupMenu(){
		popupMenu = null;
		if(null!=selectedShapeKnowledge)
			selectedShapeKnowledge.s.color = selectedShapeColor;
		
		selectedShapeKnowledge = null;
		repaint();
	}
	
	private synchronized void cachingImage(final String path){
		if(null == cachedImages.get(path)){
			new Thread(){
				@Override
				public void run(){
					Image img = Toolkit.getDefaultToolkit().createImage(path);
					cachedImages.put(path, img);
					repaint();
					//System.out.println("Image("+path+") caching COMPLETE.");
				}
			}.start();
		}
	}
	
	private java.awt.Color shapeColor2JREColor(Color color){
		switch(color){
		case BLACK:
			return java.awt.Color.BLACK;
		case CYAN:
			return java.awt.Color.CYAN;
		case RED:
			return java.awt.Color.RED;
		case YELLOW:
			return java.awt.Color.YELLOW;
		case BLUE:
			return java.awt.Color.BLUE;
		case GRAY:
			return java.awt.Color.GRAY;
		case DARK_GRAY:
			return java.awt.Color.DARK_GRAY;
		case GREEN:
			return java.awt.Color.GREEN;
		case LIGHT_GRAY:
			return java.awt.Color.LIGHT_GRAY;
		case MAGENTA:
			return java.awt.Color.MAGENTA;
		case ORANGE:
			return java.awt.Color.ORANGE;
		case PINK:
			return java.awt.Color.PINK;
		case WHITE:
			return java.awt.Color.WHITE;
		}
		return java.awt.Color.GREEN;
	}
	
	private Point2df getStringLeftBottomCoord(String str, Point2df center){
		int width = fm.stringWidth(str);
		int height = fm.getHeight();
		return new Point2df(center.x-(float)width/2.0f, center.y+(float)height/2.0f);
	}
	
	private boolean isVisibleOnScreen(Shape s){
		Point2df projectedLT = projectToScreen(s.lt);
		Point2df projectedRB = projectToScreen(s.rb);
		int width = frame.getSize().width;
		int height = frame.getSize().height;
		
		if(width<projectedLT.x || height<projectedLT.y || projectedRB.x<0 || projectedRB.y<0)
			return false;
		return true;
	}
	
	private Graphics getCurrentGraphics(){
		return g;
	}
	
	@Override
	public void mousePressed(MouseEvent e) {
		if(null != BrainmapRealWorld.sketchApp){
			Point2df point = new Point2df((float)e.getX(), (float)e.getY());
			SketchApp app = BrainmapRealWorld.sketchApp;
			
			if(null != popupMenu){
				String menu = popupMenu.selectMenu(projectToMap(point));
				if(null == menu)
					closePopupMenu();
			}
			
			else if(MouseEvent.BUTTON1 == e.getButton()){
				isRightButton = false;
				app.postEvent(new PressEvent(point, PressEvent.ACTION.PRESS, PressEvent.TYPE.LEFT));
			}
			
			else if(MouseEvent.BUTTON3 == e.getButton()){
				isRightButton = true;
				app.postEvent(new PressEvent(point, PressEvent.ACTION.PRESS, PressEvent.TYPE.RIGHT));
			}
		}
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		if(null != BrainmapRealWorld.sketchApp){
			Point2df point = new Point2df((float)e.getX(), (float)e.getY());
			SketchApp app = BrainmapRealWorld.sketchApp;
			
			if(null != popupMenu){
				String menu = popupMenu.selectMenu(projectToMap(point));
				
				if(null != menu){
					if(menu.equals("Edit Color")){
						popupMenu.showPalette();
						repaint();
					}else{
						app.postEvent(new PopupMenuEvent(menu));
						closePopupMenu();
					}
				}				
			}
			
			else if(MouseEvent.BUTTON1 == e.getButton())
				app.postEvent(new PressEvent(point, PressEvent.ACTION.RELEASE, PressEvent.TYPE.LEFT));
		
			else if(MouseEvent.BUTTON3 == e.getButton())
				app.postEvent(new PressEvent(point, PressEvent.ACTION.RELEASE, PressEvent.TYPE.RIGHT));
		}
	}
	
	@Override
	public void mouseDragged(MouseEvent e) {
		if(null != BrainmapRealWorld.sketchApp){
			Point2df point = new Point2df((float)e.getX(), (float)e.getY());
			SketchApp app = BrainmapRealWorld.sketchApp;
			
			if(null != popupMenu){}
			else if(isRightButton)	
				app.postEvent(new PressEvent(point, PressEvent.ACTION.DRAG, PressEvent.TYPE.RIGHT));
			else
				app.postEvent(new PressEvent(point, PressEvent.ACTION.DRAG, PressEvent.TYPE.LEFT));
		}
	}
	
	@Override
	public void keyTyped(KeyEvent e) {
		if(isKeyBoradEnabled){
			if(null != BrainmapRealWorld.sketchApp){
				int unicodechar = e.getKeyChar();
				
				if(unicodechar == BS)
					unicodechar = SketchApp.BS;
				
				SketchApp app = BrainmapRealWorld.sketchApp;
				app.postEvent(new cf.app.brainmap.events.KeyEvent(unicodechar));
			}
		}
	}
	
	@Override
	public void mouseWheelMoved(MouseWheelEvent e) {
		if(e.getWheelRotation() > 0){	// Zoom-in
			resolution += 0.1;
			//repaint();
		}
		
		else{	// Zoom-out
			resolution -= 0.1;
			//repaint();
		}
	}
	
	@Override	public void mouseEntered(MouseEvent e) {}
	@Override	public void mouseExited(MouseEvent e) {}
	@Override	public void mouseMoved(MouseEvent e) {}
	@Override	public void mouseClicked(MouseEvent e) {}
	@Override	public void keyPressed(KeyEvent arg0) {}
	@Override	public void keyReleased(KeyEvent arg0) {}
	@Override   public void windowActivated(WindowEvent arg0) {}

	@Override
	public void windowClosed(WindowEvent arg0) {
		//System.out.println("Window Closed");
	}

	@Override
	public void windowClosing(WindowEvent arg0) {
		//System.out.println("Window Closing");
		int ret = showMapFileDialog(MapFileEvent.TYPE.SAVE, "Do you want to Save the map?", "Save");
		
		switch(ret){
			case JOptionPane.NO_OPTION:
				hideDrawingPane();
				SketchApp app = BrainmapRealWorld.sketchApp;
				app.postEvent(new SessionEvent(SessionEvent.TYPE.EXIT, BrainmapRealWorld.brainmapApp));
				app.postEvent(new SessionEvent(SessionEvent.TYPE.SHOW, app));
				break;
		}
	}

	@Override
	public void windowDeactivated(WindowEvent arg0) {}
	public void windowDeiconified(WindowEvent arg0) {}
	public void windowIconified(WindowEvent arg0) {}
	public void windowOpened(WindowEvent arg0) {}
}
